'''
Console frame class for PyGPSClient application.

This handles a scrollable text box into which the serial data is printed.

Created on 12 Sep 2020

@author: semuadmin
'''

from threading import Event as ThreadEvent
from tkinter import Frame, Text, Scrollbar, S, E, W, END, HORIZONTAL, VERTICAL, N

from .globals import TAGS, BGCOL, FGCOL


class ConsoleFrame(Frame):
    '''
    Frame inheritance class for plotting area.
    '''

    def __init__(self, app, *args, **kwargs):
        '''
        Constructor.
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)

        Frame.__init__(self, self.__master, *args, **kwargs)

        self._body()
        self._do_layout()
        self._attach_events()

    def _body(self):
        '''
        Set up frame and widgets.
        '''

        self.option_add("*Font", self.__app.font_sm)
        self._console_fg = FGCOL
        self._console_bg = BGCOL
        self.width, self.height = self.get_size()

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.sblogv = Scrollbar(self, orient=VERTICAL)
        self.sblogh = Scrollbar(self, orient=HORIZONTAL)
        self.txt_console = Text(self,
                                bg=self._console_bg,
                                fg=self._console_fg,
                                yscrollcommand=self.sblogv.set,
                                xscrollcommand=self.sblogh.set,
                                wrap="none")
        self.sblogh.config(command=self.txt_console.xview)
        self.sblogv.config(command=self.txt_console.yview)
        # Making the textbox read only
        self.txt_console.configure(state='disabled')

    def _do_layout(self):
        '''
        Set position of widgets in frame
        '''

        self.txt_console.grid(column=0, row=0, pady=1, padx=1, sticky=(N, S, E, W))
        self.sblogv.grid(column=1, row=0, sticky=(N, S , E))
        self.sblogh.grid(column=0, row=1, sticky=(S, E , W))

    def _attach_events(self):
        '''
        Bind events to frame
        '''

        self.bind("<Configure>", self._on_resize)

    def update_console(self, data):
        '''
        Print the latest data stream to the console in raw (NMEA) or 
        parsed (key,value pair) format.

        'maxlines' defines the maximum number of scrollable lines that are 
        retained in the text box on a FIFO basis.
        '''

        settings = self.__app.frm_settings.get_settings()
        con = self.txt_console
        con.configure(state='normal')
        con.insert(END, data + '\n')

        # format of this array of tuples is (tag, highlight color)
        self._tag_line(data, TAGS)

        idx = float(con.index('end'))  # Lazy but it works
        if idx > settings['maxlines']:
            con.delete("1.0", "2.0")  # Remember these look like floats but they're not!

        con.update()
        if settings['autoscroll']:
            con.see('end')
        con.configure(state='disabled')

    def _tag_line(self, line, tags):
        '''
        Highlights any occurrence of tags in line - each tag
        must be a tuple of (search term, highlight color)
        '''

        con = self.txt_console
        idx = con.index('end-1c')
        last = int(idx[0:idx.find('.')]) - 1

        for count, tag in enumerate(tags):
            match, color = tag
            connect = line.find(match)
            end = connect + len(match)

            if connect != -1:  # If search string found in line
                con.tag_add(count, f'{last}.{connect}', f'{last}.{end}')
                con.tag_config(count, foreground=color)

    def _on_resize(self, event):
        '''
        Resize frame
        '''

        self.width, self.height = self.get_size()

    def get_size(self):
        '''
        Get current object size.
        '''

        self.update_idletasks()  # Make sure we know about any resizing
        return (self.winfo_width(), self.winfo_height())



'''
Status Bar frame class for PyGPSClient application.

This handles the status bar notifications at the foot of the window.

Created on 12 Sep 2020

@author: semuadmin
'''

from tkinter import ttk, Frame, Label, StringVar, N, S, W, E, VERTICAL


class StatusFrame(Frame):
    '''
    Frame inheritance class for _status bar.
    '''

    def __init__(self, app, *args, **kwargs):
        '''
        Constructor
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)
        Frame.__init__(self, self.__master, *args, **kwargs)

        self._status = StringVar()
        self._connection = StringVar()
        self.width, self.height = self.get_size()
        self._body()

        self.bind("<Configure>", self._on_resize)

    def _body(self):
        '''
        Set up frame and widgets.
        '''

        self.grid_rowconfigure(0, weight=1)

        self.option_add("*Font", self.__app.font_md)

        self._lbl_connection = Label(self, textvariable=self._connection, anchor=W)
        self._lbl_status = Label(self, textvariable=self._status, anchor=W)
        self._lbl_connection.grid(column=0, row=0, sticky=(W, E))
        ttk.Separator(self, orient=VERTICAL).grid(column=1, row=0, sticky=(N, S))
        self._lbl_status.grid(column=2, row=0, sticky=(W, E))

    def set_connection(self, connection, color="blue"):
        '''
        Sets connection description in status bar.
        '''

        self._lbl_connection.config(fg=color)
        self._connection.set("  " + connection)

    def set_status(self, message, color="blue"):
        '''
        Sets message in status bar.
        '''

        message = (message[:75] + '..') if len(message) > 75 else message
        self._lbl_status.config(fg=color)
        self._status.set("  " + message)

    def clear_status(self):
        '''
        Clears status bar.
        '''

        self._connection.set("")
        self._status.set("")

    def _on_resize(self, event):
        '''
        Resize frame
        '''

        self.width, self.height = self.get_size()

    def get_size(self):
        '''
        Get current frame size.
        '''

        self.update_idletasks()  # Make sure we know about any resizing
        width = self.winfo_width()
        height = self.winfo_height()
        return (width, height)



'''

Collection of global constants and helper methods

Created on 14 Sep 2020

@author: semuadmin
'''

import os
from math import sin, cos, pi
from serial import PARITY_NONE, PARITY_EVEN, PARITY_ODD, PARITY_MARK, PARITY_SPACE

DIRNAME = os.path.dirname(__file__)
ICON_APP = os.path.join(DIRNAME, 'resources/iconmonstr-location-27-32.png')
ICON_CONN = os.path.join(DIRNAME, 'resources/iconmonstr-link-8-24.png')
ICON_DISCONN = os.path.join(DIRNAME, 'resources/iconmonstr-link-10-24.png')
ICON_GO = os.path.join(DIRNAME, 'resources/iconmonstr-media-control-4-32.png')
ICON_STOP = os.path.join(DIRNAME, 'resources/iconmonstr-media-control-12-32.png')
ICON_POS = os.path.join(DIRNAME, 'resources/iconmonstr-location-1-24.png')
ICON_SEND = os.path.join(DIRNAME, 'resources/iconmonstr-arrow-12-24.png')
ICON_EXIT = os.path.join(DIRNAME, 'resources/iconmonstr-door-6-24.png')
ICON_PENDING = os.path.join(DIRNAME, 'resources/iconmonstr-time-6-24.png')
ICON_CONFIRMED = os.path.join(DIRNAME, 'resources/iconmonstr-check-mark-8-24.png')
IMG_WORLD = os.path.join(DIRNAME, 'resources/world.png')
BTN_CONNECT = "\u25b6"  # text on "Connected" button
BTN_DISCONNECT = "\u2587"  # text on "Disconnected" button

MAPURL = "https://www.mapquestapi.com/staticmap/v5/map?key={}&locations={},{}&zoom={}&defaultMarker=marker-sm-616161-ff4444&shape=radius:{}|weight:1|fill:ccffff50|border:88888850|{},{}&size={},{}"
MAP_UPDATE_INTERVAL = 60  # how frequently the mapquest api is called to update the web map
SAT_EXPIRY = 60  # how long passed satellites are kept in the sky and graph views
MAX_SNR = 60  # upper limit of graphview snr axis
DEVICE_ACCURACY = 2.5  # nominal GPS device accuracy (CEP) in meters
HDOP_RATIO = 20  # arbitrary calibration of accuracy against HDOP
KNOWNGPS = ('GPS', 'gps', 'GNSS', 'gnss', 'Garmin', 'garmin', 'U-Blox', 'u-blox')
BAUDRATES = (115200, 57600, 38400, 19200, 9600, 4800)
PARITIES = {"Even":PARITY_EVEN, "Odd": PARITY_ODD, "Mark": PARITY_MARK, "Space": PARITY_SPACE, "None": PARITY_NONE}
# serial port timeout; lower is better for app response
# but you may lose packets on high latency connections
SERIAL_TIMEOUT = .2

NMEA_PROTOCOL = 0
UBX_PROTOCOL = 1
MIXED_PROTOCOL = 2

DISCONNECTED = 0
CONNECTED = 1
NOPORTS = 2
WIDGETU1 = (250, 250)
WIDGETU2 = (350, 250)
WIDGETU3 = (950, 350)
BGCOL = "gray24"  # default widget background color
FGCOL = "white"  # default widget foreground color
ENTCOL = "azure"  # default data entry field background color
READONLY = "readonly"
ADVOFF = "\u25bc"
ADVON = "\u25b2"
ERROR = "ERR!"
DDD = "DD.D"
DMM = "DM.M"
DMS = "D.M.S"
UMM = "Metric m/s"
UMK = "Metric kmph"
UI = "Imperial mph"
UIK = "Imperial knots"

# List of tags to highlight in console
TAGS = [("DTM", "deepskyblue"), ("GBS", "pink"), \
        ("GGA", "orange"), ("GSV", "yellow"), ("GLL", "orange"), ("TXT", "lightgrey"), \
        ("GSA", "green2"), ("RMC", "orange"), ("VTG", "deepskyblue"), ("lat", "lightblue1"), \
        ("lon", "lightblue1"), ("lat_dir", "lightblue1"), ("lon_dir", "lightblue1"), \
        ("altitude", "lightblue1"), ("pdop", "lightblue1"), ("UBX", "lightblue1"), \
        ("UBX00", "aquamarine2"), ("UBX03", "yellow"), ("UBX04", "cyan"), ("UBX05", "orange"), \
        ("UBX06", "orange"), ("ZDA", "cyan"), ("GNS", "orange"), ("VLW", "deepskyblue"),
        ("GST", "mediumpurple2"), ("INF-ERROR", "red3"), ("INF-WARNING", "orange"), ("INF-NOTICE", "blue"), \
        ("vdop", "lightblue1"), ("hdop", "lightblue1"), ("h_acc", "lightblue1"), \
        ("v_acc", "lightblue1"), ("spd_over_grnd_kmph", "lightblue1"), \
        ("true_track", "lightblue1"), ("mode_fix_type", "lightblue1"), \
        ("datum", "lightblue1"), ("ACK-ACK", "green2"), ("ACK-NAK", "orange red"), \
        ("CFG-MSG", "cyan"), ("xb5b", "lightblue1"), ("NAV-SOL", "green2"), ("NAV-POSLLH", "orange"), \
        ("NAV-VELECEF", "deepskyblue"), ("NAV-VELNED", "deepskyblue"), ("NAV-SVINFO", "yellow"),
        ("NAV-TIMEUTC", "cyan"), ("NAV-STATUS", "green2"), ("NAV-PVT", "orange"),
        ("NAV-DOP", "mediumpurple2"), ("NAV-CLOCK", "cyan"), ("NAV-SBAS", "yellow"),
        ("NAV-SAT", "yellow"), ("NAV-POSECEF", "orange"), ("NAV-TIMEGLO", "cyan"),
        ("NAV-TIMEBDS", "cyan"), ("NAV-TIMEGAL", "cyan"), ("NAV-ORB", "yellow"),
        ("NAV-TIMELS", "cyan"), ("NAV-TIMEGPS", "cyan"), ("RXM", "skyblue1"), \
        ("MON", "skyblue1"), ("LOG", "skyblue1")]


def deg2rad(deg: float) -> float:
    '''
    Convert degrees to radians
    '''

    return deg * pi / 180


def cel2cart(elevation: float, azimuth: float) -> (float, float):
    '''
    Convert celestial coordinates (degrees) to Cartesian coordinates
    '''

    elevation = deg2rad(elevation)
    azimuth = deg2rad(azimuth)
    x = cos(azimuth) * cos(elevation)
    y = sin(azimuth) * cos(elevation)
    return (x, y)


def deg2dms(degrees: float, latlon: float) -> (float, float):
    '''
    Convert decimal degrees to degrees minutes seconds string
    '''

    negative = degrees < 0
    degrees = abs(degrees)
    minutes, seconds = divmod(degrees * 3600, 60)
    degrees, minutes = divmod(minutes, 60)
    if negative:
        sfx = 'S' if latlon == "lat" else 'W'
    else:
        sfx = 'N' if latlon == "lat" else 'E'
    return str(int(degrees)) + '\u00b0' + str(int(minutes)) + '\u2032' \
            +str(round(seconds, 3)) + '\u2033' + sfx


def deg2dmm(degrees: float, latlon: float) -> (float, float):
    '''
    Convert decimal degrees to degrees decimal minutes string
    '''

    negative = degrees < 0
    degrees = abs(degrees)
    degrees, minutes = divmod(degrees * 60, 60)
    if negative:
        sfx = 'S' if latlon == "lat" else 'W'
    else:
        sfx = 'N' if latlon == "lat" else 'E'
    return str(int(degrees)) + '\u00b0' + str(round(minutes, 5)) + '\u2032' + sfx


def m2ft(meters: float) -> float:
    '''
    Convert meters to feet
    '''

    return meters * 3.28084


def ft2m(feet: float) -> float:
    '''
    Convert feet to meters
    '''

    return feet / 3.28084


def kmph2mph(kmph: float) -> float:
    '''
    Convert kilometers per hour to miles per hour
    '''

    return kmph * 0.6213712419


def mph2kmph(mph: float) -> float:
    '''
    Convert miles per hour to kilometers per hour
    '''

    return mph / 0.6213712419


def kmph2ms(kmph: float) -> float:
    '''
    Convert kilometers per hour to meters per second
    '''

    return kmph * 0.2777778


def kmph2knots(kmph: float) -> float:
    '''
    Convert kilometers per hour to knots
    '''

    return kmph / 1.852001


def ms2kmph(ms: float) -> float:
    '''
    Convert meters per second to kilometers per hour
    '''

    return ms / 0.2777778


def pos2iso6709(lat :float, lon: float, alt: float, crs: str="WGS_84") -> str:
    '''
    convert decimal degrees and alt to iso6709 format
    '''

    lati = '-' if lat < 0 else '+'
    loni = '-' if lon < 0 else '+'
    alti = '-' if alt < 0 else '+'
    iso6709 = lati + str(abs(lat)) + loni + str(abs(lon)) + alti + str(alt) + "CRS" + crs + "/"
    return iso6709


def hsv2rgb(h: float, s: float, v: float) -> str:
    '''
    Convert HSV values (in range 0-1) to RGB color string.
    '''

    if s == 0.0:
        v = int(v * 255)
        return v, v, v
    i = int(h * 6.)
    f = (h * 6.) - i
    p = v * (1. - s)
    q = v * (1. - s * f)
    t = v * (1. - s * (1. - f))
    i %= 6
    if i == 0:
        r, g, b = v, t, p
    if i == 1:
        r, g, b = q, v, p
    if i == 2:
        r, g, b = p, v, t
    if i == 3:
        r, g, b = p, q, v
    if i == 4:
        r, g, b = t, p, v
    if i == 5:
        r, g, b = v, p, q

    rgb = int(r * 255), int(g * 255), int(b * 255)
    return "#%02x%02x%02x" % rgb


def snr2col(snr: int) -> str:
    '''
    Convert satellite signal-to-noise ratio to a color
    high = green, low = red 
    '''

    return hsv2rgb(snr / (MAX_SNR * 2.5), .8, .8)



'''
Filehandler class for PyGPSClient application

This handles all the file i/o

Created on 16 Sep 2020

@author: semuadmin
'''

import os
from pathlib import Path

HOME = str(Path.home())


class FileHandler():
    '''
    File handler class.
    '''

    def __init__(self, app, *args, **kwargs):
        '''
        Constructor.
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)

    def load_apikey(self) -> str:
        '''
        Load MapQuest web map api key from user's home directory.
        '''

        filepath = os.path.join(HOME, 'mqapikey')
        try:
            with open(filepath, 'r') as file:
                apikey = file.read()
        except OSError:
            # Error message will be displayed on mapview widget if invoked
            apikey = ''

        return apikey

    def load_user_presets(self) -> str:
        '''
        Load user configuration message presets from user's home directory.
        '''

        presets = []
        filepath = os.path.join(HOME, 'ubxpresets')
        try:
            with open(filepath, 'r') as file:
                for line in file:
                    presets.append(line)
        except OSError:
            presets = ''

        return presets



'''
About Dialog Box class for PyGPSClient application.

Created on 20 Sep 2020

@author: semuadmin
'''

from tkinter import Toplevel, Label, Button
from webbrowser import open_new_tab
from PIL import ImageTk, Image

from .globals import ICON_APP
from .strings import ABOUTTXT, COPYRIGHTTXT, DLGABOUT, \
                                GITHUBURL

from ._version import __version__

VERSION = __version__

class AboutDialog():
    '''
    About dialog box class
    '''

    def __init__(self, app):
        '''
        Initialise Toplevel dialog
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)
        self._dialog = Toplevel()
        self._dialog.title = DLGABOUT
        self._dialog.geometry("+%d+%d" % (self.__master.winfo_rootx() + 50,
                                          self.__master.winfo_rooty() + 50))
        self._dialog.attributes('-topmost', 'true')
        self._img_icon = ImageTk.PhotoImage(Image.open (ICON_APP))

        self._body()
        self._do_layout()
        self._attach_events()

    def _body(self):
        '''
        Set up widgets.
        '''

        self._lbl_title = Label(self._dialog, text=DLGABOUT)
        self._lbl_title.config(font=self.__app.font_md2)
        self._lbl_icon = Label(self._dialog, image=self._img_icon)
        self._lbl_desc = Label(self._dialog, text=ABOUTTXT, wraplength=250,
                               font=self.__app.font_sm)
        self._lbl_version = Label(self._dialog, text="Version: " + VERSION, font=self.__app.font_sm)
        self._lbl_copyright = Label(self._dialog, text=COPYRIGHTTXT, fg="blue",
                                    font=self.__app.font_sm,cursor="hand2")
        self._btn_ok = Button(self._dialog, text="OK", width=8, command=self.ok_press,
                              font=self.__app.font_md)


    def _do_layout(self):
        '''
        Arrange widgets in dialog.
        '''

        self._lbl_title.grid(column=0, row=0, padx=5, pady=3)
        self._lbl_icon.grid(column=0, row=1, padx=5, pady=3)
        self._lbl_desc.grid(column=0, row=2, padx=5, pady=3)
        self._lbl_version.grid(column=0, row=3, padx=5, pady=3)
        self._lbl_copyright.grid(column=0, row=4, padx=5, pady=3)
        self._btn_ok.grid(column=0, row=5, ipadx=3, ipady=3, padx=5, pady=3)

    def _attach_events(self):
        '''
        Bind events to dialog.
        '''

        self._lbl_copyright.bind("<Button-1>", lambda e: open_new_tab(GITHUBURL))
        self._btn_ok.bind("<Return>", self.ok_press)
        self._btn_ok.focus_set()

    def ok_press(self, *args, **kwargs):
        '''
        Handle OK button press.
        '''

        self.__master.update_idletasks()
        self._dialog.destroy()



'''
ENGLISH language string literals for PyGPSClient application

Created on 12 Sep 2020

@author: semuadmin
'''

TITLE = "PyGPSClient"
WIKIURL = "https://en.wikipedia.org/wiki/NMEA_0183"
GITHUBURL = "https://github.com/semuconsulting/PyGPSClient"

COPYRIGHTTXT = "\u00A9 SEMU Consulting 2020\nBSD 2 Clause License. All Rights Reserved"

INTROTXT = "Welcome to PyGPSClient!"
INTROTXTNOPORTS = "Welcome to PyGPSClient! Please connect a serial GPS device and restart."

HELPTXT = "Help..About - display About dialog."

ABOUTTXT = "PyGPSClient is a free, open-source NMEA & UBX GPS client application written" \
    +" entirely in Python and tkinter.\n\n" \
    +"Instructions and source code are available on Github at the link below."

# Message text
SETINITTXT = "Settings initialised"
VALERROR = "ERROR! Please correct highlighted entries"
SAVEERROR = "ERROR! File could not be saved to specified directory"
OPENFILEERROR = "ERROR! File could not be opened"
BADJSONERROR = "ERROR! Invalid metadata file"
NMEAVALERROR = "Value error in NMEA message: {}"
SEROPENERROR = "Error opening serial port {}"
NOWEBMAPERROR1 = "Map not available."
NOWEBMAPERROR2 = "mqapikey file not found or invalid."
NOWEBMAPERROR3 = "Unable to download web map."
NOWEBMAPERROR4 = "Check Internet connection."
NOWEBMAPERROR5 = "No satellite fix."
WEBMAPHTTPERROR = 'HTTP error downloading web map. Are you connected to the Internet?'
WEBMAPERROR = 'Error downloading web map'
SAVETITLE = "Select Save Directory"
SELTITLE = "Select File for Import"
WAITNMEADATA = "Waiting for NMEA data..."
WAITUBXDATA = "Waiting for UBX data..."
STOPDATA = "Serial reader process stopped"
NOTCONN = "Not connected"
UBXPOLL = "Polling current UBX configuration..."

# Menu text
MENUFILE = "File"
MENUVIEW = "View"
MENUOPTION = "Options"
MENUSAVE = "Save Settings"
MENULOAD = "Load Settings"
MENUEXIT = "Exit"
MENUCAN = "Cancel"
MENURST = "Reset"
MENUHIDESE = "Hide Settings"
MENUSHOWSE = "Show Settings"
MENUHIDEUBX = "Hide UBX Config"
MENUSHOWUBX = "Show UBX Config"
MENUHIDESB = "Hide Status Bar"
MENUSHOWSB = "Show Status Bar"
MENUHIDECON = "Hide Console"
MENUSHOWCON = "Show Console"
MENUHIDEMAP = "Hide Map"
MENUSHOWMAP = "Show Map"
MENUHIDESATS = "Hide Satellites"
MENUSHOWSATS = "Show Satellites"
MENUUBXCONFIG = "UBX Configuration Dialog"
MENUHOWTO = "How To"
MENUABOUT = "About"
MENUHELP = "Help"

# Button text
BTNPLOT = "PLOT"
BTNSAVE = "Save"
BTNCAN = "Cancel"
BTNRST = "Reset"

# Label text
LBLCTL = "Controls"
LBLSET = "Settings"

# Dialog text
DLGUBXCONFIG = "UBX Configuration"
DLGABOUT = "PyGPSClient"
DLGHOWTO = "How To Use PyGPSCLient"



'''
Banner frame class for PyGPSClient application.

This handles the top banner which prominently displays the current coordinates and status.

Created on 13 Sep 2020

@author: semuadmin
'''
# pylint: disable=line-too-long

from platform import system
from tkinter import Frame, Label, Button, StringVar, font, N, S, W, E, SUNKEN

from PIL import ImageTk, Image
from pyubx2 import UBXMessage

from .globals import deg2dmm, deg2dms, m2ft, kmph2mph, kmph2ms, kmph2knots, \
                                DMM, DMS, UMM, UI, UIK, ICON_CONN, ICON_DISCONN, \
                                CONNECTED, DISCONNECTED, BGCOL, FGCOL, ADVOFF, ADVON


class BannerFrame(Frame):
    '''
    Frame inheritance class for banner.
    '''

    def __init__(self, app, *args, **kwargs):
        '''
        Constructor.
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)
        Frame.__init__(self, self.__master, *args, **kwargs)

        self._time = StringVar()
        self._lat = StringVar()
        self._lon = StringVar()
        self._alt = StringVar()
        self._speed = StringVar()
        self._alt_u = StringVar()
        self._speed_u = StringVar()
        self._track = StringVar()
        self._dop = StringVar()
        self._hvdop = StringVar()
        self._hvacc = StringVar()
        self._fix = StringVar()
        self._siv = StringVar()
        self._sip = StringVar()
        self._status = False
        self._show_advanced = True

        self._bgcol = BGCOL
        self._fgcol = FGCOL
        self.config(bg=self._bgcol)
        self._img_conn = ImageTk.PhotoImage(Image.open (ICON_CONN))
        self._img_disconn = ImageTk.PhotoImage(Image.open (ICON_DISCONN))
        self.width, self.height = self.get_size()

        self._body()
        self._do_layout()
        self._attach_events()

    def _body(self):
        '''
        Set up frame and widgets.
        '''

        for i in range(2, 5):
            self.grid_columnconfigure(i, weight=1)
            self.grid_rowconfigure(0, weight=1)
            self.grid_rowconfigure(1, weight=1)
        self._frm_connect = Frame(self, bg=BGCOL)
        self._frm_toggle = Frame(self, bg=BGCOL)
        self._frm_basic = Frame(self, bg=BGCOL, relief=SUNKEN)
        self._frm_advanced = Frame(self, bg=BGCOL)

        self.option_add("*Font", self.__app.font_md2)
        self._lbl_ltime = Label(self._frm_basic, text="utc:", bg=self._bgcol, fg=self._fgcol, anchor=N)
        self._lbl_llat = Label(self._frm_basic, text="lat:", bg=self._bgcol, fg=self._fgcol, anchor=N)
        self._lbl_llon = Label(self._frm_basic, text="lon:", bg=self._bgcol, fg=self._fgcol, anchor=N)
        self._lbl_lalt = Label(self._frm_basic, text="alt:", bg=self._bgcol, fg=self._fgcol, anchor=N)
        self._lbl_lspd = Label(self._frm_basic, text="speed:", bg=self._bgcol, fg=self._fgcol, anchor=N)
        self._lbl_ltrk = Label(self._frm_basic, text="track:", bg=self._bgcol, fg=self._fgcol, anchor=N)
        self._btn_toggle = Button(self._frm_toggle, text=ADVON, width=3,
                                  command=self._toggle_advanced)
        self._lbl_lfix = Label(self._frm_basic, text="fix:", bg=self._bgcol, fg=self._fgcol, anchor=N)
        self._lbl_lsiv = Label(self._frm_advanced, text="siv:", bg=self._bgcol, fg=self._fgcol, anchor=N)
        self._lbl_lsip = Label(self._frm_advanced, text="sip:", bg=self._bgcol, fg=self._fgcol, anchor=N)
        self._lbl_lpdop = Label(self._frm_advanced, text="pdop:", bg=self._bgcol, fg=self._fgcol, anchor=N)
        self._lbl_lacc = Label(self._frm_advanced, text="acc:", bg=self._bgcol, fg=self._fgcol, anchor=N)

        self.option_add("*Font", self.__app.font_lg)
        self._lbl_status = Label(self._frm_connect, bg=self._bgcol, image=self._img_conn, fg="blue")
        self._lbl_time = Label(self._frm_basic, textvariable=self._time, bg=self._bgcol, fg="cyan")
        self._lbl_lat = Label(self._frm_basic, textvariable=self._lat, bg=self._bgcol, fg="orange")
        self._lbl_lon = Label(self._frm_basic, textvariable=self._lon, bg=self._bgcol, fg="orange")
        self._lbl_alt = Label(self._frm_basic, textvariable=self._alt, bg=self._bgcol, fg="orange")
        self._lbl_spd = Label(self._frm_basic, textvariable=self._speed, bg=self._bgcol, fg="deepskyblue")
        self._lbl_trk = Label(self._frm_basic, textvariable=self._track, bg=self._bgcol, fg="deepskyblue")
        self._lbl_fix = Label(self._frm_basic, textvariable=self._fix, bg=self._bgcol, fg="white")
        self._lbl_siv = Label(self._frm_advanced, textvariable=self._siv, bg=self._bgcol, fg="yellow")
        self._lbl_sip = Label(self._frm_advanced, textvariable=self._sip, bg=self._bgcol, fg="yellow")
        self._lbl_pdop = Label(self._frm_advanced, textvariable=self._dop, bg=self._bgcol, fg="mediumpurple2")

        self.option_add("*Font", self.__app.font_sm)
        self._lbl_lalt_u = Label(self._frm_basic, textvariable=self._alt_u, bg=self._bgcol, fg="orange",
                                anchor=S)
        self._lbl_lspd_u = Label(self._frm_basic, textvariable=self._speed_u, bg=self._bgcol, fg="deepskyblue",
                                anchor=S)
        self._lbl_hvdop = Label(self._frm_advanced, textvariable=self._hvdop, bg=self._bgcol, fg="mediumpurple2")
        self._lbl_hvacc = Label(self._frm_advanced, textvariable=self._hvacc, bg=self._bgcol, fg="aquamarine2")
        self._lbl_lacc_u = Label(self._frm_advanced, textvariable=self._alt_u, bg=self._bgcol, fg="aquamarine2", anchor=N)

    def _do_layout(self):
        '''
        Position widgets in frame.
        '''

        self._lbl_status.grid(column=0, row=0, padx=8, pady=3, sticky=W)
        self._lbl_ltime.grid(column=1, row=0, pady=3, sticky=W)
        self._lbl_time.grid(column=2, row=0, pady=3, sticky=W)
        self._lbl_llat.grid(column=3, row=0, pady=3, sticky=W)
        self._lbl_lat.grid(column=4, row=0, pady=3, sticky=W)
        self._lbl_llon.grid(column=5, row=0, pady=3, sticky=W)
        self._lbl_lon.grid(column=6, row=0, pady=3, sticky=W)
        self._lbl_lalt.grid(column=7, row=0, pady=3, sticky=W)
        self._lbl_alt.grid(column=8, row=0, pady=3, sticky=W)
        self._lbl_lalt_u.grid(column=9, row=0, pady=0, sticky=W)
        self._lbl_lspd.grid(column=10, row=0, pady=3, sticky=W)
        self._lbl_spd.grid(column=11, row=0, pady=3, sticky=W)
        self._lbl_lspd_u.grid(column=12, row=0, pady=0, sticky=W)
        self._lbl_ltrk.grid(column=13, row=0, pady=3, sticky=W)
        self._lbl_trk.grid(column=14, row=0, pady=3, sticky=W)
        self._lbl_lfix.grid(column=15, row=0, pady=3, sticky=W)
        self._lbl_fix.grid(column=16, row=0, pady=3, sticky=W)

        self._lbl_lsiv.grid(column=0, row=0, pady=3, sticky=W)
        self._lbl_siv.grid(column=1, row=0, pady=3, sticky=W)
        self._lbl_lsip.grid(column=2, row=0, pady=3, sticky=W)
        self._lbl_sip.grid(column=3, row=0, pady=3, sticky=W)
        self._lbl_lpdop.grid(column=4, row=0, pady=3, sticky=W)
        self._lbl_pdop.grid(column=5, row=0, pady=3, sticky=W)
        self._lbl_hvdop.grid(column=6, row=0, pady=0, sticky=W)
        self._lbl_lacc.grid(column=7, row=0, pady=3, sticky=W)
        self._lbl_hvacc.grid(column=8, row=0, pady=0, sticky=W)
        self._lbl_lacc_u.grid(column=9, row=0, pady=0, sticky=W)

        self._btn_toggle.grid(column=0, row=0, padx=8, pady=3, sticky=(N, E))

        self._toggle_advanced()

    def _toggle_advanced(self):
        '''
        Toggle advanced banner frame on or off.
        '''

        self._frm_connect.grid(column=0, row=0, rowspan=2,
                               pady=3, ipadx=3, ipady=3, sticky=(N, W))
        self._frm_basic.grid(column=1, row=0, pady=3, sticky=(W))
        self._frm_toggle.grid(column=5, row=0, rowspan=2, pady=3,
                              sticky=(N, E))
        self._show_advanced = not self._show_advanced
        if self._show_advanced:
            self._frm_advanced.grid(column=1, row=1, pady=3, sticky=(W))
            self._btn_toggle.config(text=ADVON)
        else:
            self._frm_advanced.grid_forget()
            self._btn_toggle.config(text=ADVOFF)

    def _attach_events(self):
        '''
        Bind events to frame.
        '''

        self.bind("<Configure>", self._on_resize)

    def update_banner(self, **kwargs):
        '''
        Sets text of banner from keyword parms.
        '''

        settings = self.__app.frm_settings.get_settings()
        disp_format = settings['format']
        units = settings['units']

        if 'time' in kwargs:
            self._time.set(kwargs['time'])
        if 'lat' in kwargs:
            if kwargs['lat'] == '':
                self._lat.set('N/A')
            else:
                if disp_format == DMS:
                    self._lat.set(deg2dms(kwargs['lat'], 'lat'))
                elif disp_format == DMM:
                    self._lat.set(deg2dmm(kwargs['lat'], 'lat'))
                else:
                    self._lat.set(round(kwargs['lat'], 5))
        if 'lon' in kwargs:
            if kwargs['lon'] == '':
                self._lon.set('N/A')
            else:
                if disp_format == DMS:
                    self._lon.set(deg2dms(kwargs['lon'], 'lon'))
                elif disp_format == DMM:
                    self._lon.set(deg2dmm(kwargs['lon'], 'lon'))
                else:
                    self._lon.set(round(kwargs['lon'], 5))
        if 'alt' in kwargs:
            if kwargs['alt'] is None:
                self._alt.set('N/A')
                self._alt_u.set('')
            else:
                if units in (UI, UIK):
                    self._alt.set(round(m2ft(float(kwargs['alt'])), 1))
                    self._alt_u.set('ft')
                else:
                    self._alt.set(round(kwargs['alt'], 1))
                    self._alt_u.set('m')
        if 'speed' in kwargs:
            if kwargs['speed'] is None:
                self._speed.set('N/A')
                self._speed_u.set('')
            else:
                if units == UI:
                    self._speed.set(round(kmph2mph(float(kwargs['speed'])), 1))
                    self._speed_u.set('mph')
                elif units == UIK:
                    self._speed.set(round(kmph2knots(float(kwargs['speed'])), 1))
                    self._speed_u.set('knots')
                elif units == UMM:
                    self._speed.set(round(kmph2ms(float(kwargs['speed'])), 1))
                    self._speed_u.set('m/s')
                else:
                    self._speed.set(round(kwargs['speed'], 1))
                    self._speed_u.set('kmph')
        if 'track' in kwargs:
            if kwargs['track'] is None:
                self._track.set('N/A')
            else:
                self._track.set(str(round(kwargs['track'], 1)))
        if 'siv' in kwargs:
            self._siv.set(str(kwargs['siv']).zfill(2))
        if 'sip' in kwargs:
            self._sip.set(str(kwargs['sip']).zfill(2))
        if 'dop' in kwargs:
            dop = kwargs['dop']
            self._dop.set(str(dop) + " " + UBXMessage.dop2str(dop))
        if 'hdop' in kwargs and 'vdop' in kwargs:
            self._hvdop.set("hdop " + str(kwargs['hdop']) + "\nvdop " + str(kwargs['vdop']))
        if 'hacc' in kwargs and 'vacc' in kwargs:
            if units in (UI, UIK):
                hacc = round(m2ft(kwargs['hacc']), 1)
                vacc = round(m2ft(kwargs['vacc']), 1)
            else:
                hacc = round(kwargs['hacc'], 1)
                vacc = round(kwargs['vacc'], 1)
            self._hvacc.set("hacc " + str(hacc) + "\nvacc " + str(vacc))
        if 'fix' in kwargs:
            if kwargs['fix'] in ('3D', '3D + DR'):
                self._lbl_fix.config(fg="green2")
            elif kwargs['fix'] in ('2D', 'DR'):
                self._lbl_fix.config(fg="orange")
            else:
                self._lbl_fix.config(fg="red")
            self._fix.set(kwargs['fix'])
        if 'status' in kwargs:
            self._status = kwargs['status']
            if self._status == CONNECTED:
                self._lbl_status.configure(image=self._img_conn)
            else:
                self._lbl_status.configure(image=self._img_disconn)

    def _set_fontsize(self):
        '''
        Adjust font sizes according to frame size
        '''

        w = self.width
        # Cater for slightly different font behaviour on Linux
        if system() in ("W32", "Darwin"):
            val = 55
            lbl = 75
            sup = 85
        else:
            val = 70
            lbl = 85
            sup = 95

        sz = min(int(w / val), 18)
        self._lbl_status.config(font=font.Font(size=sz))
        self._lbl_time.config(font=font.Font(size=sz))
        self._lbl_lat.config(font=font.Font(size=sz))
        self._lbl_lon.config(font=font.Font(size=sz))
        self._lbl_alt.config(font=font.Font(size=sz))
        self._lbl_spd.config(font=font.Font(size=sz))
        self._lbl_trk.config(font=font.Font(size=sz))
        self._lbl_pdop.config(font=font.Font(size=sz))
        self._lbl_fix.config(font=font.Font(size=sz))
        self._lbl_sip.config(font=font.Font(size=sz))
        self._lbl_siv.config(font=font.Font(size=sz))

        sz = min(int(w / lbl), 14)
        self._lbl_ltime.config(font=font.Font(size=sz))
        self._lbl_llat.config(font=font.Font(size=sz))
        self._lbl_llon.config(font=font.Font(size=sz))
        self._lbl_lalt.config(font=font.Font(size=sz))
        self._lbl_lspd.config(font=font.Font(size=sz))
        self._lbl_ltrk.config(font=font.Font(size=sz))
        self._lbl_lpdop.config(font=font.Font(size=sz))
        self._lbl_lfix.config(font=font.Font(size=sz))
        self._lbl_lsip.config(font=font.Font(size=sz))
        self._lbl_lsiv.config(font=font.Font(size=sz))
        self._lbl_lacc.config(font=font.Font(size=sz))

        sz = min(int(w / sup), 12)
        self._lbl_lalt_u.config(font=font.Font(size=sz))
        self._lbl_lspd_u.config(font=font.Font(size=sz))
        self._lbl_hvdop.config(font=font.Font(size=sz))
        self._lbl_hvacc.config(font=font.Font(size=sz))

    def _on_resize(self, event):
        '''
        Resize frame
        '''

        self.width, self.height = self.get_size()
        self._set_fontsize()

    def get_size(self):
        '''
        Get current frame size.
        '''

        self.update_idletasks()  # Make sure we know about any resizing
        width = self.winfo_width()
        height = self.winfo_height()
        return (width, height)



'''
UBX Protocol handler

Uses pyubx2 library for parsing

Created on 30 Sep 2020

@author: semuadmin
'''
# pylint: disable=invalid-name

from pyubx2 import UBXMessage, POLL, UBX_MSGIDS

CFG_MSG_OFF = b'\x00\x00\x00\x00\x00\x01'
CFG_MSG_ON = b'\x01\x01\x01\x01\x01\x01'
BOTH = 3
UBX = 1
NMEA = 2


class UBXHandler():
    '''
    UBXHandler class
    '''

    def __init__(self, app):
        '''
        Constructor.
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)

        self._raw_data = None
        self._parsed_data = None
        self.gsv_data = []  # Holds array of current satellites in view from NMEA GSV sentences
        self.lon = 0
        self.lat = 0
        self.alt = 0
        self.track = 0
        self.speed = 0
        self.pdop = 0
        self.hdop = 0
        self.vdop = 0
        self.hacc = 0
        self.vacc = 0
        self.utc = ''
        self.sip = 0
        self.fix = '-'
        self.ubx_baudrate = 9600
        self.ubx_inprot = 7
        self.ubx_outprot = 3

    @staticmethod
    def poll_ubx_config(serial):
        '''
        POLL current UBX device configuration (port protocols
        and message rates).

        NB: The responses and acknowledgements to these polls
        may take several seconds to arrive, particularly in
        heavy traffic.
        '''

        for msgtype in ('CFG-PRT', 'CFG-USB'):
            msg = UBXMessage('CFG', msgtype, None, POLL)
            serial.write(msg.serialize())

#         for payload in UBX_CONFIG_MESSAGES:
#             msg = UBXMessage('CFG', 'CFG-MSG', payload, POLL)
#             serial.write(msg.serialize())
#
#         for payload in (b'\x00', b'\x01'):  # UBX & NMEA
#             msg = UBXMessage('CFG', 'CFG-INF', payload, POLL)
#             serial.write(msg.serialize())

    def process_data(self, data: bytes) -> UBXMessage:
        '''
        Process UBX message type
        '''

        parsed_data = UBXMessage.parse(data, False)

        if parsed_data.identity == 'ACK-ACK':
            self._process_ACK_ACK(parsed_data)
        if parsed_data.identity == 'ACK-NAK':
            self._process_ACK_NAK(parsed_data)
        if parsed_data.identity == 'CFG-MSG':
            self._process_CFG_MSG(parsed_data)
        if parsed_data.identity == 'CFG-PRT':
            self._process_CFG_PRT(parsed_data)
        if parsed_data.identity == 'CFG-INF':
            self._process_CFG_INF(parsed_data)
        if parsed_data.identity == 'NAV-POSLLH':
            self._process_NAV_POSLLH(parsed_data)
        if parsed_data.identity == 'NAV-PVT':
            self._process_NAV_PVT(parsed_data)
        if parsed_data.identity == 'NAV-VELNED':
            self._process_NAV_VELNED(parsed_data)
        if parsed_data.identity == 'NAV-SVINFO':
            self._process_NAV_SVINFO(parsed_data)
        if parsed_data.identity == 'NAV-SOL':
            self._process_NAV_SOL(parsed_data)
        if parsed_data.identity == 'NAV-DOP':
            self._process_NAV_DOP(parsed_data)
        if data or parsed_data:
            self._update_console(data, parsed_data)

        return parsed_data

    def _update_console(self, raw_data, parsed_data):
        '''
        Write the incoming data to the console in raw or parsed format.
        '''

        if self.__app.frm_settings.get_settings()['raw']:
            self.__app.frm_console.update_console(str(raw_data))
        else:
            self.__app.frm_console.update_console(str(parsed_data))

    def _process_ACK_ACK(self, data: UBXMessage):
        '''
        Process CFG-MSG sentence - UBX message configuration.
        '''

        msgtype = UBX_MSGIDS[data.clsID + data.msgID]

        # update the UBX config panel
        if self.__app.dlg_ubxconfig is not None:
            self.__app.dlg_ubxconfig.update('ACK-ACK', msgtype=msgtype)

    def _process_ACK_NAK(self, data: UBXMessage):
        '''
        Process CFG-MSG sentence - UBX message configuration.
        '''

        msgtype = UBX_MSGIDS[data.clsID + data.msgID]

        # update the UBX config panel
        if self.__app.dlg_ubxconfig is not None:
            self.__app.dlg_ubxconfig.update('ACK-NAK', msgtype=msgtype)

    def _process_CFG_MSG(self, data: UBXMessage):
        '''
        Process CFG-MSG sentence - UBX message configuration.
        '''

        msgtype = UBX_MSGIDS[data.msgClass + data.msgID]

        # update the UBX config panel
        if self.__app.dlg_ubxconfig is not None:
            self.__app.dlg_ubxconfig.update('CFG-MSG', msgtype=msgtype)

    def _process_CFG_INF(self, data: UBXMessage):
        '''
        Process CFG-INF sentence - UBX info message configuration.
        '''

        # update the UBX config panel
        if self.__app.dlg_ubxconfig is not None:
            self.__app.dlg_ubxconfig.update('CFG-INF')

    def _process_CFG_PRT(self, data: UBXMessage):
        '''
        Process CFG-PRT sentence - UBX port configuration.
        '''

        self.ubx_baudrate = data.baudRate
        self.ubx_inprot = int.from_bytes(data.inProtoMask, 'little', signed=False)
        self.ubx_outprot = int.from_bytes(data.outProtoMask, 'little', signed=False)

        # update the UBX config panel
        if self.__app.dlg_ubxconfig is not None:
            self.__app.dlg_ubxconfig.update('CFG-PRT', baudrate=self.ubx_baudrate,
                                            inprot=self.ubx_inprot,
                                            outprot=self.ubx_outprot)

    def _process_NAV_POSLLH(self, data: UBXMessage):
        '''
        Process NAV-LLH sentence - Latitude, Longitude, Height.
        '''

        try:
            self.utc = UBXMessage.itow2utc(data.iTOW)
            self.lat = data.lat / 10 ** 7
            self.lon = data.lon / 10 ** 7
            self.alt = data.hMSL / 1000
            self.hacc = data.hAcc / 1000
            self.vacc = data.vAcc / 1000
            self.__app.frm_banner.update_banner(time=self.utc, lat=self.lat,
                                                lon=self.lon, alt=self.alt,
                                                hacc=self.hacc, vacc=self.vacc)

            if self.__app.frm_settings.get_settings()['webmap']:
                self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, hacc=self.hacc,
                                                    vacc=self.vacc, fix='3D', static=False)
            else:
                self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, hacc=self.hacc,
                                                    vacc=self.vacc, fix='3D', static=True)
        except ValueError:
            # self.__app.set_status(ube.UBXMessageError(err), "red")
            pass

    def _process_NAV_PVT(self, data: UBXMessage):
        '''
        Process NAV-PVT sentence -  Navigation position velocity time solution
        '''

        try:
            self.utc = UBXMessage.itow2utc(data.iTOW)
            self.lat = data.lat / 10 ** 7
            self.lon = data.lon / 10 ** 7
            self.alt = data.hMSL / 1000
            self.hacc = data.hAcc / 1000
            self.vacc = data.vAcc / 1000
            self.pdop = data.pDOP / 100
            self.sip = data.numSV
            self.speed = data.gSpeed / 100
            self.track = data.headMot / 10 ** 5
            fix = UBXMessage.gpsfix2str(data.fixType)
            self.__app.frm_banner.update_banner(time=self.utc, lat=self.lat,
                                                lon=self.lon, alt=self.alt,
                                                hacc=self.hacc, vacc=self.vacc,
                                                dop=self.pdop, sip=self.sip,
                                                speed=self.speed, fix=fix,
                                                track=self.track)

            if self.__app.frm_settings.get_settings()['webmap']:
                self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, hacc=self.hacc,
                                                  vacc=self.vacc, fix='3D', static=False)
            else:
                self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, hacc=self.hacc,
                                                  vacc=self.vacc, fix='3D', static=True)
        except ValueError:
            # self.__app.set_status(ube.UBXMessageError(err), "red")
            pass

    def _process_NAV_VELNED(self, data: UBXMessage):
        '''
        Process NAV-VELNED sentence - Velocity Solution in North East Down format.
        '''

        try:
            self.track = data.heading / 10 ** 5
            self.speed = data.gSpeed / 100
            self.__app.frm_banner.update_banner(speed=self.speed, track=self.track)
        except ValueError:
            # self.__app.set_status(ube.UBXMessageError(err), "red")
            pass

    def _process_NAV_SVINFO(self, data: UBXMessage):
        '''
        Process NAV-SVINFO sentences - Space Vehicle Information.
        '''

        try:
            self.gsv_data = []
            num_siv = int(data.numCh)
            self.__app.frm_banner.update_banner(siv=num_siv)

            for i in range(num_siv):
                idx = "_{0:0=2d}".format(i + 1)
                # TODO is there an easier/better way to do this without exec()?:
                exec("self.gsv_data.append((data.svid" + str(idx) + \
                     ", data.elev" + str(idx) + \
                     ", data.azim" + str(idx) + ", data.cno" + str(idx) + "))")
            self.__app.frm_satview.update_sats(self.gsv_data)
            self.__app.frm_graphview.update_graph(self.gsv_data, num_siv)
        except ValueError:
            # self.__app.set_status(ube.UBXMessageError(err), "red")
            pass

    def _process_NAV_SOL(self, data: UBXMessage):
        '''
        Process NAV-SOL sentence - Navigation Solution.
        '''

        try:
            self.pdop = data.pDOP / 100
            self.sip = data.numSV
            fix = UBXMessage.gpsfix2str(data.gpsFix)

            self.__app.frm_banner.update_banner(dop=self.pdop, fix=fix, sip=self.sip)
        except ValueError:
            # self.__app.set_status(ube.UBXMessageError(err), "red")
            pass

    def _process_NAV_DOP(self, data: UBXMessage):
        '''
        Process NAV-DOP sentence - Dilution of Precision.
        '''

        try:
            self.pdop = data.pDOP / 100
            self.hdop = data.hDOP / 100
            self.vdop = data.vDOP / 100

            self.__app.frm_banner.update_banner(dop=self.pdop, hdop=self.hdop, vdop=self.vdop)
        except ValueError:
            # self.__app.set_status(ube.UBXMessageError(err), "red")
            pass



'''
PyGPSClient - Main tkinter application class

Created on 12 Sep 2020

@author: semuadmin
'''

from tkinter import Tk, Frame, N, S, E, W, PhotoImage, font

from .about_dialog import AboutDialog
from .banner_frame import BannerFrame
from .console_frame import ConsoleFrame
from .filehandler import FileHandler
from .globals import ICON_APP
from .graphview_frame import GraphviewFrame
from .map_frame import MapviewFrame
from .menu_bar import MenuBar
from .serial_handler import SerialHandler
from .settings_frame import SettingsFrame
from .skyview_frame import SkyviewFrame
from .status_frame import StatusFrame
from .strings import TITLE, INTROTXT, MENUHIDESE, MENUSHOWSE, \
                                MENUHIDESB, MENUSHOWSB, MENUHIDECON, MENUSHOWCON, MENUHIDEMAP, \
                                MENUSHOWMAP, MENUHIDESATS, MENUSHOWSATS, INTROTXTNOPORTS
from .ubx_config_dialog import UBXConfigDialog
from .nmea_handler import NMEAHandler
from .ubx_handler import UBXHandler

from ._version import __version__

VERSION = __version__


class App(Frame):
    '''
    Main GUI Application Class
    '''

    def __init__(self, master, *args, **kwargs):
        '''
        Set up main application and add frames
        '''

        self.__master = master

        Frame.__init__(self, self.__master, *args, **kwargs)

        self.__master.protocol('WM_DELETE_WINDOW', self.exit)
        self.__master.title(TITLE)
        self.__master.iconphoto(True, PhotoImage(file=ICON_APP))
        self._show_settings = False  # Flag to toggle settings frame
        self._show_ubxconfig = True  # Flag to toggle pyubx2 config frame
        self._show_status = False  # Flag to toggle status bar
        self._show_console = False  # Flag to toggler console
        self._show_map = False  # Flag to toggle status bar
        self._show_sats = False  # Flag to toggler console

        # Instantiate protocol handler classes
        self.file_handler = FileHandler(self)
        self.serial_handler = SerialHandler(self)
        self.nmea_handler = NMEAHandler(self)
        self.ubx_handler = UBXHandler(self)
        self.dlg_ubxconfig = None

        # Load web map api key if there is one
        self.api_key = self.file_handler.load_apikey()

        self._body()
        self._do_layout()
        self._attach_events()

        # Initialise widgets
        self.frm_satview.init_sats()
        self.frm_graphview.init_graph()
        self.frm_banner.update_banner(status=False)

    def _body(self):
        '''
        Set up frame and widgets
        '''

        for i in range(3):
            self.__master.grid_columnconfigure(i, weight=1)
        self.__master.grid_rowconfigure(1, weight=1)
        self.__master.grid_rowconfigure(2, weight=1)
        self._set_default_fonts()

        self.frm_status = StatusFrame(self, borderwidth=2, relief="groove")
        self.frm_banner = BannerFrame(self, borderwidth=2, relief="groove")
        self.frm_settings = SettingsFrame(self, borderwidth=2, relief="groove")
        self.frm_console = ConsoleFrame(self, borderwidth=2, relief="groove")
        self.frm_mapview = MapviewFrame(self, borderwidth=2, relief="groove")
        self.frm_satview = SkyviewFrame(self, borderwidth=2, relief="groove")
        self.frm_graphview = GraphviewFrame(self, borderwidth=2, relief="groove")
        self.menu = MenuBar(self)

        self.__master.config(menu=self.menu)

    def _do_layout(self):
        '''
        Arrange widgets in main application frame
        '''

        self.frm_banner.grid(column=0, row=0, columnspan=5, padx=2, pady=2,
                             sticky=(N, S, E, W))
        self.toggle_status()
        self.toggle_settings()
        self.toggle_console()
        self.toggle_map()
        self.toggle_sats()

        if self.frm_settings.get_settings()['noports']:
            self.set_status(INTROTXTNOPORTS, "red")
        else:
            self.set_status(INTROTXT, "blue")
        # self.frm_settings._lbx_port.focus_set()

    def _attach_events(self):
        '''
        Bind events to main application
        '''

        self.__master.bind('<<ubx_read>>', self.serial_handler.on_read)
        self.__master.bind_all("<Control-q>", self.exit)

    def _set_default_fonts(self):
        '''
        Set default fonts for entire application
        '''
        # pylint: disable=attribute-defined-outside-init

        self.font_vsm = font.Font(size=8)
        self.font_sm = font.Font(size=10)
        self.font_md = font.Font(size=12)
        self.font_md2 = font.Font(size=14)
        self.font_lg = font.Font(size=18)

    def toggle_settings(self):
        '''
        Toggle Settings Frame on or off
        '''

        if self._show_settings:
            self.frm_settings.grid_forget()
            self._show_settings = False
            self.menu.view_menu.entryconfig(0, label=MENUSHOWSE)
        else:
            self.frm_settings.grid(column=4, row=1, rowspan=2, padx=2,
                                   pady=2, sticky=(N, W, E))
            self._show_settings = True
            self.menu.view_menu.entryconfig(0, label=MENUHIDESE)

    def toggle_status(self):
        '''
        Toggle Status Bar on or off
        '''

        if self._show_status:
            self.frm_status.grid_forget()
            self._show_status = False
            self.menu.view_menu.entryconfig(1, label=MENUSHOWSB)
        else:
            self.frm_status.grid(column=0, row=3, columnspan=5, padx=2,
                                 pady=2, sticky=(W, E))
            self._show_status = True
            self.menu.view_menu.entryconfig(1, label=MENUHIDESB)

    def toggle_console(self):
        '''
        Toggle Console frame on or off
        '''

        if self._show_console:
            self.frm_console.grid_forget()
            self._show_console = False
            self.menu.view_menu.entryconfig(2, label=MENUSHOWCON)
        else:
            self.frm_console.grid(column=0, row=1, columnspan=3, padx=2,
                                  pady=2, sticky=(N, S, E, W))
            self._show_console = True
            self.menu.view_menu.entryconfig(2, label=MENUHIDECON)

    def toggle_map(self):
        '''
        Toggle Map frame on or off
        '''

        if self._show_map:
            self.frm_mapview.grid_forget()
            self._show_map = False
            self.menu.view_menu.entryconfig(3, label=MENUSHOWMAP)
        else:
            self.frm_mapview.grid(column=2, row=2, padx=2, pady=2,
                                  sticky=(N, S, E, W))
            self._show_map = True
            self.menu.view_menu.entryconfig(3, label=MENUHIDEMAP)

    def toggle_sats(self):
        '''
        Toggle Satview and Graphview frames on or off
        '''

        if self._show_sats:
            self.frm_satview.grid_forget()
            self.frm_graphview.grid_forget()
            self._show_sats = False
            self.menu.view_menu.entryconfig(4, label=MENUSHOWSATS)
        else:

            self.frm_satview.grid(column=0, row=2, padx=2, pady=2,
                                  sticky=(N, S, E, W))
            self.frm_graphview.grid(column=1, row=2, padx=2, pady=2,
                                    sticky=(N, S, E, W))
            self._show_sats = True
            self.menu.view_menu.entryconfig(4, label=MENUHIDESATS)

    def set_connection(self, message, color="blue"):
        '''
        Sets connection description in status bar
        '''

        self.frm_status.set_connection(message, color)

    def set_status(self, message, color="black"):
        '''
        Sets text of status bar
        '''

        self.frm_status.set_status(message, color)

    def about(self):
        '''
        Open About dialog
        '''

        AboutDialog(self)

    def ubxconfig(self):
        '''
        Open UBX Config dialog
        Update idle tasks and flush input buffer first
        to minimise blocking on startup during heavy
        traffic
        '''

        self.__master.update_idletasks()
        self.serial_handler.flush()
        self.dlg_ubxconfig = UBXConfigDialog(self)

    def get_master(self):
        '''
        Returns application master (Tk)
        '''

        return self.__master

    def exit(self, *args, **kwargs):  # pylint: disable=unused-argument
        '''
        Kill any running processes and quit application
        '''

        self.serial_handler.stop_read_thread()
        self.serial_handler.disconnect()
        self.__master.destroy()


if __name__ == "__main__":
    ROOT = Tk()
    APP = App(ROOT)
    ROOT.mainloop()



'''
Graphview frame class for PyGPSClient application.

This handles a frame containing a graph of current satellite reception.

Created on 14 Sep 2020

@author: semuadmin
'''

from tkinter import Frame, Canvas, font, BOTH, YES

from .globals import snr2col, WIDGETU2, BGCOL, FGCOL, MAX_SNR

# Relative offsets of graph axes
AXIS_XL = 19
AXIS_XR = 10
AXIS_Y = 18


class GraphviewFrame(Frame):
    '''
    Frame inheritance class for plotting satellite view.
    '''

    def __init__(self, app, *args, **kwargs):
        '''
        Constructor.
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)

        Frame.__init__(self, self.__master, *args, **kwargs)

        def_w, def_h = WIDGETU2
        self.width = kwargs.get('width', def_w)
        self.height = kwargs.get('height', def_h)
        self._bgcol = BGCOL
        self._fgcol = FGCOL
        self._body()

        self.bind("<Configure>", self._on_resize)

    def _body(self):
        '''
        Set up frame and widgets.
        '''

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.can_graphview = Canvas(self, width=self.width, height=self.height, bg=self._bgcol)
        self.can_graphview.pack(fill=BOTH, expand=YES)

    def init_graph(self):
        '''
        Initialise graph view
        '''

        w, h = self.width, self.height
        resize_font = font.Font(size=min(int(h / 25), 10))
        ticks = int(MAX_SNR / 10)
        self.can_graphview.delete("all")
        self.can_graphview.create_line(AXIS_XL, 5, AXIS_XL , h - AXIS_Y, fill=self._fgcol)
        self.can_graphview.create_line(w - AXIS_XR + 2, 5 , w - AXIS_XR + 2, h - AXIS_Y,
                                       fill=self._fgcol)
        for i in range(ticks, 0, -1):
            y = (h - AXIS_Y) * i / ticks
            self.can_graphview.create_line(AXIS_XL, y, w - AXIS_XR + 2, y, fill=self._fgcol)
            self.can_graphview.create_text(10, y, text=str(MAX_SNR - (i * 10)), angle=90,
                                           fill=self._fgcol, font=resize_font)
#         self.can_graphview.create_text(10, (h - AXIS_Y - 1) / 2, text="dbHz", angle=90,
#                                        fill=self._fgcol, font=resize_font)

    def update_graph(self, data, siv=16):
        '''
        Plot satellites' elevation and azimuth position.
        Automatically adjust y axis according to number of satellites in view.
        '''

        if siv == 0:
            return

        w, h = self.width, self.height
        self.init_graph()

        offset = AXIS_XL + 1
        colwidth = (w - AXIS_XL - AXIS_XR) / siv
        resize_font = font.Font(size=min(int(colwidth / 2), 10))
        for d in sorted(data):  # sort by ascending prn
            prn, _, _, snr = d
            if snr in ('', '0', 0):
                snr = 1
            else:
                snr = int(snr)
            snr_y = int(snr) * (h - AXIS_Y - 1) / MAX_SNR
            if int(prn) > 96:  # OTHER e.g. GAL
                ol_col = "grey"
            elif 65 <= int(prn) <= 96:  # GLONASS
                ol_col = "brown"
            elif 33 <= int(prn) <= 64:  # SBAS
                ol_col = "blue"
            else:  # original GPS
                ol_col = "black"
            prn = f"{int(prn):02}"
            self.can_graphview.create_rectangle(offset, h - AXIS_Y - 1, offset + colwidth,
                                                h - AXIS_Y - 1 - snr_y, outline=ol_col,
                                                fill=snr2col(snr))
            self.can_graphview.create_text(offset + colwidth / 2, h - 10, text=prn,
                                           fill=self._fgcol, font=resize_font, angle=35)
            offset += colwidth

        self.can_graphview.update()

    def _on_resize(self, event):
        '''
        Resize frame
        '''

        self.width, self.height = self.get_size()

    def get_size(self):
        '''
        Get current canvas size.
        '''

        self.update_idletasks()  # Make sure we know about any resizing
        width = self.can_graphview.winfo_width()
        height = self.can_graphview.winfo_height()
        return (width, height)



'''
UBX Configuration frame class for PyGPSClient application.

This handles the configuration of a U-BLox GPS device via the UBX protocol.

Created on 22 Sep 2020

@author: semuadmin
'''

from tkinter import ttk, Toplevel, Frame, Checkbutton, Radiobutton, Listbox, \
                    Spinbox, Scrollbar, messagebox, Button, Label, StringVar, \
                    IntVar, N, S, E, W, LEFT, VERTICAL

from PIL import ImageTk, Image
from pyubx2 import UBXMessage, POLL, SET, UBX_CONFIG_MESSAGES

from .globals import BGCOL, FGCOL, ENTCOL, ICON_APP, ICON_SEND, ICON_EXIT, \
                     ICON_PENDING, ICON_CONFIRMED, BAUDRATES, READONLY
from .strings import DLGUBXCONFIG
from .ubx_handler import UBXHandler as ubh, CFG_MSG_OFF, CFG_MSG_ON

MSG_PRESETS = {
'CFG-CFG - RESTORE FACTORY DEFAULTS': 'CFG-CFG',
'CFG-CFG - Store current configuration': 'CFG-CFG',
'CFG-MSG - Turn ON minimum NMEA msgs': 'CFG-MSG',
'CFG-MSG - Turn ON all NMEA msgs': 'CFG-MSG',
'CFG-MSG - Turn OFF all NMEA msgs': 'CFG-MSG',
'CFG-MSG - Turn ON minimum UBX NAV msgs': 'CFG-MSG',
'CFG-MSG - Turn ON all UBX NAV msgs': 'CFG-MSG',
'CFG-MSG - Turn OFF all UBX NAV msgs': 'CFG-MSG',
'CFG-INF - Turn ON all INF msgs': 'CFG-INF',
'CFG-INF - Turn OFF all non-error INF msgs': 'CFG-INF',
'CFG-MSG - Turn ON all LOG msgs': 'CFG-MSG',
'CFG-MSG - Turn OFF all LOG msgs': 'CFG-MSG',
'CFG-MSG - Turn ON all MON msgs': 'CFG-MSG',
'CFG-MSG - Turn OFF all MON msgs': 'CFG-MSG',
'CFG-MSG - Turn ON all RXM msgs': 'CFG-MSG',
'CFG-MSG - Turn OFF all RXM msgs': 'CFG-MSG',
'CFG-PRT - Poll Port config': 'CFG-PRT',
'CFG-INF - Poll Info message config': 'CFG:INF',
}


class UBXConfigDialog():
    ''',
    Frame inheritance class for application settings and controls.
    '''

    def __init__(self, app, *args, **kwargs):
        '''
        Constructor.
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)
        self._dialog = Toplevel()
        self._dialog.transient(self.__app)
        self._dialog.resizable(False, False)
        self._dialog.title = DLGUBXCONFIG
        wd, hd = self.get_size()
        wa = self.__master.winfo_width()
        ha = self.__master.winfo_height()
        self._dialog.geometry("+%d+%d" % (int(wa / 2 - wd / 2), int(ha / 2 - hd / 2)))
#         self._dialog.attributes('true')
        self._img_icon = ImageTk.PhotoImage(Image.open (ICON_APP))
        self._img_pending = ImageTk.PhotoImage(Image.open(ICON_PENDING))
        self._img_confirmed = ImageTk.PhotoImage(Image.open(ICON_CONFIRMED))
        self._img_send = ImageTk.PhotoImage(Image.open(ICON_SEND))
        self._img_exit = ImageTk.PhotoImage(Image.open(ICON_EXIT))
        self._awaiting_ack = False

        # Poll current UBX configuration
        if self.__app.serial_handler.connected:
            ubh.poll_ubx_config(self.__app.serial_handler.serial)

        # Load user presets if there are any
        self._userpresets = self.__app.file_handler.load_user_presets()

        #  Initialise up key variables
        self._preset_command = None
        self._status = StringVar()
        self._ubx_baudrate = StringVar()
        self._ubx_inprot = IntVar()
        self._ubx_outprot = IntVar()
        self._dtm_state = IntVar()
        self._gbs_state = IntVar()
        self._gga_state = IntVar()
        self._gll_state = IntVar()
        self._gsa_state = IntVar()
        self._gsv_state = IntVar()
        self._rmc_state = IntVar()
        self._txt_state = IntVar()
        self._vtg_state = IntVar()
        self._zda_state = IntVar()
        self._xxx_state = IntVar()
        self._ubx00_state = IntVar()
        self._ubx03_state = IntVar()
        self._ubx04_state = IntVar()
        self._ubx05_state = IntVar()
        self._ubx06_state = IntVar()
        self._navaopstatus_state = IntVar()
        self._navdop_state = IntVar()
        self._navposllh_state = IntVar()
        self._navpvt_state = IntVar()
        self._navsol_state = IntVar()
        self._navsvinfo_state = IntVar()
        self._navsbas_state = IntVar()
        self._navtimeutc_state = IntVar()
        self._navvelned_state = IntVar()

#         self._cfg_msg_states = {
#         'DTM': self._dtm_state,
#         'GBS': self._gbs_state,
#         'GGA': self._gga_state,
#         'GLL': self._gll_state,
#         'GSA': self._gsa_state,
#         'GSV': self._gsv_state,
#         'RMC': self._rmc_state,
#         'TXT': self._txt_state,
#         'VTG': self._vtg_state,
#         'ZDA': self._zda_state,
#         'XXX': self._xxx_state,  # for error testing
#         'UBX00': self._ubx00_state,
#         'UBX03': self._ubx03_state,
#         'UBX04': self._ubx04_state,
#         'UBX05': self._ubx05_state,
#         'UBX06': self._ubx06_state,
#         'NAV-AOPSTATUS': self._navaopstatus_state,
#         'NAV-DOP': self._navdop_state,
#         'NAV-POSLLH': self._navposllh_state,
#         'NAV-PVT': self._navpvt_state,
#         'NAV-SOL': self._navsol_state,
#         'NAV-SVINFO': self._navsvinfo_state,
#         'NAV-SBAS': self._navsbas_state,
#         'NAV-TIMEUTC': self._navtimeutc_state,
#         'NAV-VELNED': self._navvelned_state}

        self._body()
        self._do_layout()
        self._attach_events()
        self._reset()

    def _body(self):
        '''
        Set up frame and widgets.
        '''

        LBL_COL = "snow2"
        self._frm_container = Frame(self._dialog, borderwidth=2, relief="groove")
        con = self._frm_container
        for i in range(7):
            con.grid_columnconfigure(i, weight=1)
        con.grid_rowconfigure(1, weight=1)

        con.option_add("*Font", self.__app.font_sm)

        self._lbl_title = Label(con, text="UBX Configuration", bg=BGCOL, fg=FGCOL,
                                justify=LEFT, font=self.__app.font_md)
        self._lbl_cfg_port = Label(con, text="CFG-PRT Port and Protocol Configuration",
                                   bg=LBL_COL, anchor='w')
        self._lbl_ubx_baudrate = Label(con, text="Baud rate")
        self._spn_ubx_baudrate = Spinbox(con,
                                         values=(BAUDRATES),
                                         width=8, state=READONLY, readonlybackground=ENTCOL,
                                         wrap=True, textvariable=self._ubx_baudrate)
        self._lbl_inprot = Label(con, text="Input")
        self._rad_inprot_nmea = Radiobutton(con, text="NMEA",
                                             variable=self._ubx_inprot, value=2)
        self._rad_inprot_ubx = Radiobutton(con, text="UBX",
                                            variable=self._ubx_inprot, value=1)
        self._rad_inprot_rtcm = Radiobutton(con, text="RTCM",
                                            variable=self._ubx_inprot, value=4)
        self._rad_inprot_both = Radiobutton(con, text="ALL",
                                             variable=self._ubx_inprot, value=7)
        self._lbl_outprot = Label(con, text="Output")
        self._rad_outprot_nmea = Radiobutton(con, text="NMEA",
                                             variable=self._ubx_outprot, value=2)
        self._rad_outprot_ubx = Radiobutton(con, text="UBX",
                                            variable=self._ubx_outprot, value=1)
        self._rad_outprot_both = Radiobutton(con, text="ALL",
                                             variable=self._ubx_outprot, value=3)
        self._lbl_send_port = Label(con)
        self._btn_send_port = Button(con, image=self._img_send, width=50,
                                     command=self._on_send_port,
                                     font=self.__app.font_md)

        self._lbl_cfg_msg = Label(con, text="CFG-MSG - NMEA Message Filters",
                                  bg=LBL_COL, anchor='w')
        self._chk_dtm = Checkbutton(con, text="DTM",
                                    command=lambda: self._on_cfg_msg(msg=b'\xF0\x0A',
                                                                     var=self._dtm_state),
                                    variable=self._dtm_state)
        self._chk_gbs = Checkbutton(con, text="GBS",
                                    command=lambda: self._on_cfg_msg(msg=b'\xF0\x09',
                                                                     var=self._gbs_state),
                                    variable=self._gbs_state)
        self._chk_gll = Checkbutton(con, text="GLL",
                                    command=lambda: self._on_cfg_msg(msg=b'\xF0\x01',
                                                                     var=self._gll_state),
                                    variable=self._gll_state)
        self._chk_gga = Checkbutton(con, text="GGA",
                                    command=lambda: self._on_cfg_msg(msg=b'\xF0\x00',
                                                                     var=self._gga_state),
                                    variable=self._gga_state)
        self._chk_gsa = Checkbutton(con, text="GSA",
                                    command=lambda: self._on_cfg_msg(msg=b'\xF0\x02',
                                                                     var=self._gsa_state),
                                    variable=self._gsa_state)
        self._chk_gsv = Checkbutton(con, text="GSV",
                                    command=lambda: self._on_cfg_msg(msg=b'\xF0\x03',
                                                                     var=self._gsv_state),
                                    variable=self._gsv_state)
        self._chk_rmc = Checkbutton(con, text="RMC",
                                    command=lambda: self._on_cfg_msg(msg=b'\xF0\x04',
                                                                     var=self._rmc_state),
                                    variable=self._rmc_state)
        self._chk_vtg = Checkbutton(con, text="VTG",
                                    command=lambda: self._on_cfg_msg(msg=b'\xF0\x05',
                                                                     var=self._vtg_state),
                                    variable=self._vtg_state)
        self._chk_txt = Checkbutton(con, text="TXT",
                                    command=lambda: self._on_cfg_msg(msg=b'\xF0\x41',
                                                                     var=self._txt_state),
                                    variable=self._txt_state)
        self._chk_zda = Checkbutton(con, text="ZDA",
                                    command=lambda: self._on_cfg_msg(msg=b'\xF0\x08',
                                                                     var=self._zda_state),
                                    variable=self._zda_state)
        self._chk_ubx00 = Checkbutton(con, text="UBX,00",
                                      command=lambda: self._on_cfg_msg(msg=b'\xF1\x00',
                                                                       var=self._ubx00_state),
                                      variable=self._ubx00_state)
        self._chk_ubx03 = Checkbutton(con, text="UBX,03",
                                      command=lambda: self._on_cfg_msg(msg=b'\xF1\x03',
                                                                       var=self._ubx03_state),
                                      variable=self._ubx03_state)
        self._chk_ubx04 = Checkbutton(con, text="UBX,04",
                                      command=lambda: self._on_cfg_msg(msg=b'\xF1\x04',
                                                                       var=self._ubx04_state),
                                      variable=self._ubx04_state)
        self._lbl_ubx_msg = Label(con, text="CFG-MSG - UBX NAV Message Filters",
                                  bg=LBL_COL, anchor='w')
        self._chk_navaopstatus = Checkbutton(con, text="AOPSTATUS",
                                    command=lambda: self._on_cfg_msg(msg=b'\x01\x03',
                                                                     var=self._navaopstatus_state),
                                    variable=self._navaopstatus_state)
        self._chk_navsol = Checkbutton(con, text="SOL",
                                    command=lambda: self._on_cfg_msg(msg=b'\x01\x06',
                                                                     var=self._navsol_state),
                                    variable=self._navsol_state)
        self._chk_navdop = Checkbutton(con, text="DOP",
                                    command=lambda: self._on_cfg_msg(msg=b'\x01\x04',
                                                                     var=self._navdop_state),
                                    variable=self._navdop_state)
        self._chk_navposllh = Checkbutton(con, text="POSLLH",
                                    command=lambda: self._on_cfg_msg(msg=b'\x01\x02',
                                                                     var=self._navposllh_state),
                                    variable=self._navposllh_state)
        self._chk_navpvt = Checkbutton(con, text="PVT",
                                    command=lambda: self._on_cfg_msg(msg=b'\x01\x07',
                                                                     var=self._navpvt_state),
                                    variable=self._navpvt_state)
        self._chk_navsvinfo = Checkbutton(con, text="SVINFO",
                                    command=lambda: self._on_cfg_msg(msg=b'\x01\x30',
                                                                     var=self._navsvinfo_state),
                                    variable=self._navsvinfo_state)
        self._chk_navsbas = Checkbutton(con, text="SBAS",
                                      command=lambda: self._on_cfg_msg(msg=b'\x01\x32',
                                                                       var=self._navsbas_state),
                                      variable=self._navsbas_state)
        self._chk_navvelned = Checkbutton(con, text="VELNED",
                                      command=lambda: self._on_cfg_msg(msg=b'\x01\x12',
                                                                       var=self._navvelned_state),
                                      variable=self._navvelned_state)
        self._chk_navtimeutc = Checkbutton(con, text="TIMEUTC",
                                      command=lambda: self._on_cfg_msg(msg=b'\x01\x21',
                                                                       var=self._navtimeutc_state),
                                      variable=self._navtimeutc_state)

        self._lbl_presets = Label(con, text="Preset UBX Configuration Commands",
                                  bg=LBL_COL, anchor='w')
        self._lbl_preset = Label(con, text="Preset")
        self._lbx_preset = Listbox(con, border=2,
                                 relief="sunken", bg=ENTCOL,
                                 height=7, justify=LEFT,
                                 exportselection=False)
        self._scr_preset = Scrollbar(con, orient=VERTICAL)
        self._lbx_preset.config(yscrollcommand=self._scr_preset.set)
        self._scr_preset.config(command=self._lbx_preset.yview)

        self._lbl_send_preset = Label(con)
        self._btn_send_preset = Button(con, image=self._img_send, width=50, fg="green",
                                       command=self._on_send_preset, font=self.__app.font_md)
        self._lbl_status = Label(con, textvariable=self._status, anchor='w')
        self._btn_exit = Button(con, image=self._img_exit, width=50, fg="red",
                                command=self._on_exit, font=self.__app.font_md)

    def _do_layout(self):
        '''
        Position widgets in frame.
        '''

        self._frm_container.grid(column=0, row=0, columnspan=6, padx=3, pady=3,
                                 ipadx=5, ipady=5, sticky=(N, S, W, E))
        self._lbl_title.grid(column=0, row=0, columnspan=6, ipady=3, sticky=(W, E))
        self._lbl_cfg_port.grid(column=0, row=1, columnspan=6, padx=3, sticky=(W, E))
        self._lbl_ubx_baudrate.grid(column=0, row=2, columnspan=3, sticky=(W))
        self._spn_ubx_baudrate.grid(column=1, row=2, columnspan=3, sticky=(W))
        self._lbl_inprot.grid(column=0, row=3, sticky=(W))
        self._rad_inprot_nmea.grid(column=1, row=3, sticky=(W))
        self._rad_inprot_ubx.grid(column=2, row=3, sticky=(W))
        self._rad_inprot_rtcm.grid(column=3, row=3, sticky=(W))
        self._rad_inprot_both.grid(column=4, row=3, sticky=(W))
        self._lbl_outprot.grid(column=0, row=4, sticky=(W))
        self._rad_outprot_nmea.grid(column=1, row=4, sticky=(W))
        self._rad_outprot_ubx.grid(column=2, row=4, sticky=(W))
        self._rad_outprot_both.grid(column=3, row=4, sticky=(W))
        self._btn_send_port.grid(column=4, row=2, ipadx=3,
                                 ipady=3, sticky=(E))
        self._lbl_send_port.grid(column=5, row=2, ipadx=3,
                                 ipady=3, sticky=(W))

        ttk.Separator(self._frm_container).grid(column=0, row=5, columnspan=6,
                                                padx=3, pady=3, sticky=(W, E))
        self._lbl_cfg_msg.grid(column=0, row=6, columnspan=6, padx=3, sticky=(W, E))
        self._chk_dtm.grid(column=0, row=7, sticky=(W))
        self._chk_gbs.grid(column=1, row=7, sticky=(W))
        self._chk_gga.grid(column=2, row=7, sticky=(W))
        self._chk_gll.grid(column=3, row=7, sticky=(W))
        self._chk_gsa.grid(column=4, row=7, sticky=(W))
        self._chk_gsv.grid(column=5, row=7, sticky=(W))
        self._chk_rmc.grid(column=0, row=8, sticky=(W))
        self._chk_txt.grid(column=1, row=8, sticky=(W))
        self._chk_vtg.grid(column=2, row=8, sticky=(W))
        self._chk_zda.grid(column=3, row=8, sticky=(W))
        self._chk_ubx00.grid(column=0, row=9, sticky=(W))
        self._chk_ubx03.grid(column=1, row=9, sticky=(W))
        self._chk_ubx04.grid(column=2, row=9, sticky=(W))

        ttk.Separator(self._frm_container).grid(column=0, row=10, columnspan=6,
                                                padx=3, pady=3, sticky=(W, E))
        self._lbl_ubx_msg.grid(column=0, row=11, columnspan=6, padx=3, sticky=(W, E))
        self._chk_navaopstatus.grid(column=0, row=12, sticky=(W))
        self._chk_navdop.grid(column=1, row=12, sticky=(W))
        self._chk_navposllh.grid(column=2, row=12, sticky=(W))
        self._chk_navpvt.grid(column=3, row=12, sticky=(W))
        self._chk_navsbas.grid(column=4, row=12, sticky=(W))
        self._chk_navsol.grid(column=5, row=12, sticky=(W))
        self._chk_navsvinfo.grid(column=0, row=13, sticky=(W))
        self._chk_navtimeutc.grid(column=1, row=13, sticky=(W))
        self._chk_navvelned.grid(column=2, row=13, sticky=(W))

        ttk.Separator(self._frm_container).grid(column=0, row=14, columnspan=6,
                                                padx=3, pady=3, sticky=(W, E))
        self._lbl_presets.grid(column=0, row=15, columnspan=6, padx=3, sticky=(W, E))
        self._lbx_preset.grid(column=0, row=16, columnspan=4, padx=3, pady=3, sticky=(W, E))
        self._scr_preset.grid(column=3, row=16, sticky=(N, S, E))
        self._btn_send_preset.grid(column=4, row=16, ipadx=3, ipady=3,
                                   sticky=(E))
        self._lbl_send_preset.grid(column=5, row=16, rowspan=2, ipadx=3,
                                 ipady=3, sticky=(W))

        ttk.Separator(self._frm_container).grid(column=0, row=18, columnspan=6,
                                                padx=3, pady=3, sticky=(W, E))

        self._lbl_status.grid(column=0, row=19, columnspan=4, ipadx=3, ipady=3,
                            sticky=(W, E))
        self._btn_exit.grid(column=4, row=19, ipadx=3, ipady=3,
                            sticky=(E))

    def _attach_events(self):
        '''
        Bind events
        '''

        self._lbx_preset.bind("<<ListboxSelect>>", self._on_select_preset)

    def update(self, cfgtype='CFG-MSG', **kwargs):
        '''
        Update panel with latest polled configuration
        '''

        if 'baudrate' in kwargs:
            self._ubx_baudrate.set(str(kwargs['baudrate']))
        if 'inprot' in kwargs:
            self._ubx_inprot.set(kwargs['inprot'])
        if 'outprot' in kwargs:
            self._ubx_outprot.set(kwargs['outprot'])
        self._lbl_send_port.config(image=self._img_confirmed)
#         self._lbl_send_preset.config(image=self._img_confirmed)
        if self._awaiting_ack and cfgtype == 'ACK-ACK':
            self.set_status(f"{cfgtype} acknowledgement received", "green")
            self._awaiting_ack = False
        if self._awaiting_ack and cfgtype == 'ACK-NAK':
            self.set_status(f"{cfgtype} rejection received", "red")
            self._awaiting_ack = False
        elif not self._awaiting_ack and cfgtype not in ('ACK-ACK', 'ACK-NAK'):
            self.set_status(f"{cfgtype} response received", "green")

    def _reset(self):
        '''
        Reset settings to defaults.
        '''

        self._ubx_baudrate.set(str(self.__app.ubx_handler.ubx_baudrate))
        self._ubx_inprot.set(self.__app.ubx_handler.ubx_inprot)
        self._ubx_outprot.set(self.__app.ubx_handler.ubx_outprot)
        self._gbs_state.set(False)
        self._dtm_state.set(False)
        self._gbs_state.set(False)
        self._gll_state.set(True)
        self._gga_state.set(True)
        self._gsa_state.set(True)
        self._gsv_state.set(True)
        self._rmc_state.set(True)
        self._vtg_state.set(True)
        self._txt_state.set(True)
        self._zda_state.set(False)
        self._ubx00_state.set(False)
        self._ubx03_state.set(False)
        self._ubx04_state.set(False)

        idx = 0
        for pre in MSG_PRESETS:
            idx += 1
            self._lbx_preset.insert(idx, pre)
        for userpre in self._userpresets:
            idx += 1
            self._lbx_preset.insert(idx, userpre)

    def _on_cfg_msg(self, *args, **kwargs):
        '''
        Send CFG_MSG message to turn individual NMEA message types on or off
        '''

        msg = kwargs['msg']
        state = kwargs['var'].get()
        if state:  # if checkbutton checked
            onoff = CFG_MSG_ON
        else:
            onoff = CFG_MSG_OFF
        msg_payload = msg + onoff
        data = UBXMessage('CFG', 'CFG-MSG', msg_payload, 1).serialize()
        self.__app.serial_handler.serial_write(data)

    def _on_send_port(self, *args, **kwargs):
        '''
        Handle Send port config button press.
        '''

        portID = b'\x03'
        reserved0 = b'\x00'
        reserved4 = b'\x00\00'
        reserved5 = b'\x00\00'
        txReady = b'\x00\x00'
        mode = b'\x00\x00\x00\x00'
        baudRate = int.to_bytes(int(self._ubx_baudrate.get()), 4, 'little', signed=False)
        inProtoMask = int.to_bytes(self._ubx_inprot.get(), 2, 'little', signed=False)
        outProtoMask = int.to_bytes(self._ubx_outprot.get(), 2, 'little', signed=False)
        payload = portID + reserved0 + txReady + mode + baudRate + inProtoMask \
                  +outProtoMask + reserved4 + reserved5
        msg = UBXMessage('CFG', 'CFG-PRT', payload, SET)
        self.__app.serial_handler.serial_write(msg.serialize())

        self._do_poll_prt()  # poll for confirmation
        self._lbl_send_port.config(image=self._img_pending)
        self.set_status("CFG-PRT configuration message(s) sent")

    def _on_select_preset(self, *args, **kwargs):
        '''
        Preset command has been selected.
        '''

        idx = self._lbx_preset.curselection()
        self._preset_command = self._lbx_preset.get(idx)

    def _on_send_preset(self, *args, **kwargs):
        '''
        Handle Send preset button press.
        '''

        if self._preset_command == 'CFG-CFG - RESTORE FACTORY DEFAULTS':
            self._do_factory_reset()
        elif self._preset_command == 'CFG-INF - Turn ON all INF msgs':
            self._do_set_inf(True)
        elif self._preset_command == 'CFG-INF - Turn OFF all non-error INF msgs':
            self._do_set_inf(False)
        elif self._preset_command == 'CFG-MSG - Turn ON all LOG msgs':
            self._do_set_log(True)
        elif self._preset_command == 'CFG-MSG - Turn OFF all LOG msgs':
            self._do_set_log(False)
        elif self._preset_command == 'CFG-MSG - Turn ON all MON msgs':
            self._do_set_mon(True)
        elif self._preset_command == 'CFG-MSG - Turn OFF all MON msgs':
            self._do_set_mon(False)
        elif self._preset_command == 'CFG-MSG - Turn ON all RXM msgs':
            self._do_set_rxm(True)
        elif self._preset_command == 'CFG-MSG - Turn OFF all RXM msgs':
            self._do_set_rxm(False)
        elif self._preset_command == 'CFG-MSG - Turn ON minimum NMEA msgs':
            self._do_set_minnmea()
        elif self._preset_command == 'CFG-MSG - Turn ON all NMEA msgs':
            self._do_set_allnmea(True)
        elif self._preset_command == 'CFG-MSG - Turn OFF all NMEA msgs':
            self._do_set_allnmea(False)
        elif self._preset_command == 'CFG-MSG - Turn ON minimum UBX NAV msgs':
            self._do_set_minNAV()
        elif self._preset_command == 'CFG-MSG - Turn ON all UBX NAV msgs':
            self._do_set_allNAV(True)
        elif self._preset_command == 'CFG-MSG - Turn OFF all UBX NAV msgs':
            self._do_set_allNAV(False)
        elif self._preset_command == 'CFG-PRT - Poll Port config':
            self._do_poll_prt()
        elif self._preset_command == 'CFG-INF - Poll Info message config':
            self._do_poll_inf()
        else:
            messagebox.showwarning("UBX Configuration Presets", "Sorry - not yet implemented!\n\nWatch this space.",)

        self.set_status("Configuration message(s) sent")
        self._awaiting_ack = True
#         self._lbl_send_port.config(image=self._img_pending)

    def _do_poll_inf(self):
        '''
        Poll INF message configuration
        '''

        for payload in (b'\x00', b'\x01'):  # UBX & NMEA
            msg = UBXMessage('CFG', 'CFG-INF', payload, POLL)
            self.__app.serial_handler.serial_write(msg.serialize())

    def _do_poll_prt(self):
        '''
        Poll PRT message configuration
        '''

        msg = UBXMessage('CFG', 'CFG-PRT', None, POLL)
        self.__app.serial_handler.serial_write(msg.serialize())

    def _do_set_inf(self, onoff):
        '''
        Turn on device information messages INF
        '''

        if onoff:
            mask = b'\x31'  # all INF msgs
        else:
            mask = b'\x01'  # errors only
        for protocolID in (b'\x00', b'\x01'):  # UBX and NMEA
            reserved1 = b'\x00\x00\x00'
            infMsgMaskDDC = mask
            infMsgMaskUART1 = mask
            infMsgMaskUART2 = mask
            infMsgMaskUSB = mask
            infMsgMaskSPI = mask
            reserved2 = b'\x00'
            payload = protocolID + reserved1 + infMsgMaskDDC + \
                      infMsgMaskUART1 + infMsgMaskUART2 + infMsgMaskUSB + \
                      infMsgMaskSPI + reserved2
            msg = UBXMessage('CFG', 'CFG-INF', payload, SET)
            self.__app.serial_handler.serial_write(msg.serialize())
            self._do_poll_inf()  # poll results

    def _do_set_log(self, onoff):
        '''
        Turn on all device logging messages LOG
        '''

        for msgtype in UBX_CONFIG_MESSAGES:
            if msgtype[0:1] == b'\x21':
                self._do_cfgmsg(msgtype, onoff)

    def _do_set_mon(self, onoff):
        '''
        Turn on all device monitoring messages MON
        '''

        for msgtype in UBX_CONFIG_MESSAGES:
            if msgtype[0:1] == b'\x0A':
                self._do_cfgmsg(msgtype, onoff)

    def _do_set_rxm(self, onoff):
        '''
        Turn on all device receiver management messages RXM
        '''

        for msgtype in UBX_CONFIG_MESSAGES:
            if msgtype[0:1] == b'\x02':
                self._do_cfgmsg(msgtype, onoff)

    def _do_set_minnmea(self):
        '''
        Turn on minimum set of NMEA messages (GGA & GSA & GSV)
        '''

        for msgtype in UBX_CONFIG_MESSAGES:
            if msgtype[0:1] in (b'\xF0'):  # standard NMEA
                if msgtype in (b'\xF0\x00', b'\xF0\x02', b'\xF0\x03'):  # GGA, GSA, GSV
                    self._do_cfgmsg(msgtype, True)
                else:
                    self._do_cfgmsg(msgtype, False)
            if msgtype[0:1] in (b'\xF1'):  # proprietary NMEA
                self._do_cfgmsg(msgtype, False)

    def _do_set_minNAV(self):
        '''
        Turn on minimum set of UBX-NAV messages (PVT & SVINFO)
        '''

        for msgtype in UBX_CONFIG_MESSAGES:
            if msgtype[0:1] in (b'\x01'):  # UBX-NAV
                if msgtype in (b'\x01\x07', b'\x01\x30'):  # NAV-PVT, NAV-SVINFO
                    self._do_cfgmsg(msgtype, True)
                else:
                    self._do_cfgmsg(msgtype, False)

    def _do_set_allnmea(self, onoff):
        '''
        Turn on all NMEA messages
        '''

        for msgtype in UBX_CONFIG_MESSAGES:
            if msgtype[0:1] not in (b'\x0A', b'\x01'):
                self._do_cfgmsg(msgtype, onoff)

    def _do_set_allNAV(self, onoff):
        '''
        Turn on all UBX-NAV messages
        '''

        for msgtype in UBX_CONFIG_MESSAGES:
            if msgtype[0:1] == b'\x01':
                self._do_cfgmsg(msgtype, onoff)

    def _do_cfgmsg(self, msgtype, onoff):
        '''
        Send CFG-MSG for specified message type
        '''

        if onoff:
            rate = CFG_MSG_ON
        else:
            rate = CFG_MSG_OFF
        payload = msgtype + rate
        msg = UBXMessage('CFG', 'CFG-MSG', payload, SET)
        self.__app.serial_handler.serial_write(msg.serialize())

    def _do_factory_reset(self):
        '''
        Restore to factory defaults
        '''

        clearMask = b'\x07\x00\x00\x00'
        saveMask = b'\x00\x00\x00\x00'
        loadMask = b'\x07\x00\x00\x00'
        devicerMask = b'\x01'
        payload = clearMask + saveMask + loadMask + devicerMask
        msg = UBXMessage('CFG', 'CFG-CFG', payload, SET)
        self.__app.serial_handler.serial_write(msg.serialize())

    def set_status(self, message, color="blue"):
        '''
        Set status message.
        '''

        message = (message[:50] + '..') if len(message) > 50 else message
        self._lbl_status.config(fg=color)
        self._status.set("  " + message)

    def _on_exit(self, *args, **kwargs):
        '''
        Handle Exit button press.
        '''

        self.__master.update_idletasks()
        self.__app.dlg_ubxconfig = None
        self._dialog.destroy()

    def get_size(self):
        '''
        Get current frame size.
        '''

        self.__master.update_idletasks()  # Make sure we know about any resizing
        return (self._dialog.winfo_width(), self._dialog.winfo_height())



'''
Entry point for PyGPSClient Application.

Created on 12 Sep 2020

@author: semuadmin
'''

from tkinter import Tk
from pygpsclient.app import App

if __name__ == "__main__":
    ROOT = Tk()
    APP = App(ROOT)
    ROOT.mainloop()



'''
SerialHandler class for PyGPSClient application

This handles all the serial i/o , threaded read process and direction to
the appropriate protocol handler

Created on 16 Sep 2020

@author: semuadmin
'''

from io import BufferedRWPair
from threading import Thread

from serial import Serial, SerialException, SerialTimeoutException
import pyubx2.ubxtypes_core as ubt

from .globals import CONNECTED, DISCONNECTED, SERIAL_TIMEOUT, \
                                NMEA_PROTOCOL, MIXED_PROTOCOL, UBX_PROTOCOL, PARITIES
from .strings import WAITUBXDATA, STOPDATA, NOTCONN, SEROPENERROR


class SerialHandler():
    '''
    Serial handler class.
    '''

    def __init__(self, app):
        '''
        Constructor.
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)

        self._serial_object = None
        self._serial_buffer = None
        self._serial_thread = None
        self._connected = False
        self._port = None
        self._reading = False

    def __del__(self):
        '''
        Destructor.
        '''

        if self._serial_thread is not None:
            self._reading = False
            self._serial_thread = None

    def connect(self):
        '''
        Open serial connection.
        '''

        settings = self.__app.frm_settings.get_settings()

        self._port = settings['port']
        port_desc = settings['port_desc']
        baudrate = settings['baudrate']
        databits = settings['databits']
        stopbits = settings['stopbits']
        parity = PARITIES[settings['parity']]
        xonxoff = settings['xonxoff']
        rtscts = settings['rtscts']

        try:
            self._serial_object = Serial(self._port,
                                         baudrate,
                                         bytesize=databits,
                                         stopbits=stopbits,
                                         parity=parity,
                                         xonxoff=xonxoff,
                                         rtscts=rtscts,
                                         timeout=SERIAL_TIMEOUT)
            self._serial_buffer = BufferedRWPair(self._serial_object, self._serial_object)
            self.__app.frm_banner.update_banner(status=True)
            self.__app.set_connection(f"{self._port}:{port_desc} @ {str(baudrate)}", "green")
            self.__app.frm_settings.set_controls(CONNECTED)
            self._connected = True
            self.start_read_thread()
        except (SerialException, SerialTimeoutException) as err:
            self._connected = False
            self.__app.set_connection(f"{self._port}:{port_desc} @ {str(baudrate)}", "red")
            self.__app.set_status(SEROPENERROR.format(err), "red")
            self.__app.frm_banner.update_banner(status=False)
            self.__app.frm_settings.set_controls(DISCONNECTED)

    def disconnect(self):
        '''
        Close serial connection.
        '''

        if self._connected:
            try:
                self._reading = False
                self._serial_object.close()
                self.__app.frm_banner.update_banner(status=False)
                self.__app.set_connection(NOTCONN, "red")
                self.__app.set_status("", "blue")
            except (SerialException, SerialTimeoutException):
                pass

        self._connected = False
        self.__app.frm_settings.set_controls(self._connected)

    @property
    def port(self):
        '''
        Getter for port
        '''

        return self._port

    @property
    def connected(self):
        '''
        Getter for connection status
        '''

        return self._connected

    @property
    def serial(self):
        '''
        Getter for serial object
        '''

        return self._serial_object

    @property
    def buffer(self):
        '''
        Getter for serial buffer
        '''

        return self._serial_buffer

    @property
    def thread(self):
        '''
        Getter for serial thread
        '''

        return self._serial_thread

    def serial_write(self, data: bytes):
        '''
        Write binary data to serial port.
        '''

        try:
            self._serial_object.write(data)
#             self._serial_buffer.write(data)
        except (SerialException, SerialTimeoutException) as err:
            print(f"Error writing to serial port {err}")

    def start_read_thread(self):
        '''
        Start the serial reader thread.
        '''

        if self._connected:
            self._reading = True
            self.__app.set_status(WAITUBXDATA, "blue")
            self.__app.frm_mapview.reset_map_refresh()
            self._serial_object.reset_input_buffer()
            self._serial_thread = Thread(target=self._read_thread, daemon=True)
            self._serial_thread.start()

    def stop_read_thread(self):
        '''
        Stop serial reader thread.
        '''

        if self._serial_thread is not None:
            self._reading = False
            self._serial_thread = None
            self.__app.set_status(STOPDATA, "red")

    def _read_thread(self):
        '''
        THREADED PROCESS
        Reads binary data from serial port and generates virtual event to
        trigger data parsing and widget updates.
        '''

        while self._reading and self._serial_object:
            # print(f"Bytes in buffer: {self._serial_object.in_waiting}")
            if self._serial_object.in_waiting:
                self.__master.event_generate('<<ubx_read>>')

    def on_read(self, event):  # pylint: disable=unused-argument
        '''
        Action on <<ubx_read>> event - read any data in the buffer.
        '''

        if self._reading and self._serial_object is not None:
            self.__app.set_status("",)
            self._parse_data(self._serial_buffer)

    def _parse_data(self, ser: Serial) -> object:
        '''
        Read the binary data and direct to the appropriate
        UBX and/or NMEA protocol handler, depending on which protocols
        are filtered.
        '''

        parsing = True
        raw_data = None
        byte1 = ser.read(2)  # read first two bytes to determine protocol

        while parsing:
#             self.__app.update_idletasks()
            filt = self.__app.frm_settings.get_settings()['protocol']
            # if it's a UBX message (b'\b5\x62')
            if byte1 == ubt.UBX_HDR and filt in (UBX_PROTOCOL, MIXED_PROTOCOL):
#                 print(f"doing ubx {ser.peek()}")
                byten = ser.read(4)
                clsid = byten[0:1]
                msgid = byten[1:2]
                lenb = byten[2:4]
                leni = int.from_bytes(lenb, 'little', signed=False)
                byten = ser.read(leni + 2)
                plb = byten[0:leni]
                cksum = byten[leni:leni + 2]
                raw_data = ubt.UBX_HDR + clsid + msgid + lenb + plb + cksum
                self.__app.ubx_handler.process_data(raw_data)
                parsing = False
            # if it's an NMEA message ('$G' or '$P')
            elif byte1 in (b'\x24\x47', b'\x24\x50') and filt in (NMEA_PROTOCOL, MIXED_PROTOCOL):
#                 print(f"doing nmea {ser.peek()}")
                try:
                    raw_data = byte1 + ser.readline()
                    self.__app.nmea_handler.process_data(raw_data.decode("utf-8"))
                except UnicodeDecodeError:
                    continue
                parsing = False
            # else drop it like it's hot
            else:
#                 print(f"dropping {ser.peek()}")
                parsing = False

    def flush(self):
        '''
        Flush input buffer
        '''

        if self._serial_buffer is not None:
            self._serial_buffer.flush()
        if self._serial_object is not None:
            self._serial_object.flushInput()



'''
Settings frame class for PyGPSClient application.

This handles the settings/configuration panel.

Created on 12 Sep 2020

@author: semuadmin
'''

from tkinter import ttk, Frame, Button, Label, Spinbox, Scrollbar, Listbox, Scale, \
                    Checkbutton, Radiobutton, StringVar, IntVar, DoubleVar, N, S, E, \
                    W, LEFT, NORMAL, DISABLED, VERTICAL, HORIZONTAL

from PIL import ImageTk, Image
from serial.tools.list_ports import comports

from .globals import ENTCOL, DDD, DMM, DMS, UMM, UMK, UI, UIK, ADVON, \
                                ADVOFF, READONLY, CONNECTED, DISCONNECTED, NOPORTS, \
                                KNOWNGPS, ICON_CONN, ICON_DISCONN, BAUDRATES, \
                                NMEA_PROTOCOL, UBX_PROTOCOL, MIXED_PROTOCOL


class SettingsFrame(Frame):
    '''
    Frame inheritance class for application settings and controls.
    '''

    def __init__(self, app, *args, **kwargs):
        '''
        Constructor.
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)
        Frame.__init__(self, self.__master, *args, **kwargs)

        self._show_advanced = False
        self._settings = {}
        self._ports = ()
        self._port = StringVar()
        self._port_desc = StringVar()
        self._baudrate = IntVar()
        self._databits = IntVar()
        self._stopbits = DoubleVar()
        self._parity = StringVar()
        self._rtscts = IntVar()
        self._xonxoff = IntVar()
        self._protocol = IntVar()
        self._raw = IntVar()
        self._autoscroll = IntVar()
        self._maxlines = IntVar()
        self._webmap = IntVar()
        self._mapzoom = IntVar()
        self._units = StringVar()
        self._format = StringVar()
        self._noports = True
        self._validsettings = True
        self._img_conn = ImageTk.PhotoImage(Image.open(ICON_CONN))
        self._img_disconn = ImageTk.PhotoImage(Image.open(ICON_DISCONN))

        self._body()
        self._do_layout()
        self._get_ports()
        self._reset()

    def _body(self):
        '''
        Set up frame and widgets.
        '''

        for i in range(4):
            self.grid_columnconfigure(i, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.option_add("*Font", self.__app.font_sm)
        # Serial port settings
        self._frm_basic = Frame(self)
        self._lbl_port = Label(self._frm_basic, text="Port")
        self._lbx_port = Listbox(self._frm_basic, border=2,
                                 relief="sunken", bg=ENTCOL,
                                 width=28, height=5, justify=LEFT,
                                 exportselection=False)
        self._scr_portv = Scrollbar(self._frm_basic, orient=VERTICAL)
        self._scr_porth = Scrollbar(self._frm_basic, orient=HORIZONTAL)
        self._lbx_port.config(yscrollcommand=self._scr_portv.set)
        self._lbx_port.config(xscrollcommand=self._scr_porth.set)
        self._scr_portv.config(command=self._lbx_port.yview)
        self._scr_porth.config(command=self._lbx_port.xview)
        self._lbx_port.bind("<<ListboxSelect>>", self._on_select_port)
        self._lbl_baudrate = Label(self._frm_basic, text="Baud rate")
        self._spn_baudrate = Spinbox(self._frm_basic,
                                     values=(BAUDRATES),
                                     width=8, state=READONLY, readonlybackground=ENTCOL,
                                     wrap=True, textvariable=self._baudrate)
        self._btn_toggle = Button(self._frm_basic, text=ADVOFF, width=3,
                                  command=self._toggle_advanced)

        self._frm_advanced = Frame(self)
        self._lbl_databits = Label(self._frm_advanced, text="Data Bits")
        self._spn_databits = Spinbox(self._frm_advanced, values=(8, 7, 6, 5),
                                     width=3, state=READONLY, readonlybackground=ENTCOL,
                                     wrap=True, textvariable=self._databits)
        self._lbl_stopbits = Label(self._frm_advanced, text="Stop Bits")
        self._spn_stopbits = Spinbox(self._frm_advanced, values=(2, 1.5, 1),
                                     width=3, state=READONLY, readonlybackground=ENTCOL,
                                     wrap=True, textvariable=self._stopbits)
        self._lbl_parity = Label(self._frm_advanced, text="Parity")
        self._spn_parity = Spinbox(self._frm_advanced,
                                   values=("None", "Even", "Odd", "Mark", "Space"),
                                   width=6, state=READONLY, readonlybackground=ENTCOL,
                                   wrap=True, textvariable=self._parity)
        self._chk_rts = Checkbutton(self._frm_advanced, text="RTS/CTS",
                                    variable=self._rtscts)
        self._chk_xon = Checkbutton(self._frm_advanced, text="Xon/Xoff",
                                    variable=self._xonxoff)

        self._frm_buttons = Frame(self)
        self._btn_connect = Button(self._frm_buttons, width=45, height=35,
                                  image=self._img_conn,
                                  command=lambda: self.__app.serial_handler.connect())
        self._btn_disconnect = Button(self._frm_buttons, width=45, height=35,
                                     image=self._img_disconn,
                                     command=lambda: self.__app.serial_handler.disconnect(),
                                     state=DISABLED)
        self._lbl_status = Label(self._frm_buttons, font=self.__app.font_md2, text='')

        # Other configuration options
        self._frm_options = Frame(self)
        self._lbl_protocol = Label(self._frm_options, text="Protocol")
        self._rad_nmea = Radiobutton(self._frm_options, text="NMEA",
                                    variable=self._protocol, value=NMEA_PROTOCOL)
        self._rad_ubx = Radiobutton(self._frm_options, text="UBX",
                                    variable=self._protocol, value=UBX_PROTOCOL)
        self._rad_all = Radiobutton(self._frm_options, text="ALL",
                                    variable=self._protocol, value=MIXED_PROTOCOL)
        self._lbl_display = Label(self._frm_options, text="Console Display")
        self._rad_parsed = Radiobutton(self._frm_options, text="Parsed",
                                    variable=self._raw, value=0)
        self._rad_raw = Radiobutton(self._frm_options, text="Raw",
                                   variable=self._raw, value=1)
        self._lbl_format = Label(self._frm_options, text="Degrees Format")
        self._spn_format = Spinbox(self._frm_options,
                                  values=(DDD, DMS, DMM),
                                  width=6, state=READONLY, readonlybackground=ENTCOL,
                                  wrap=True, textvariable=self._format)
        self._lbl_units = Label(self._frm_options, text="Units")
        self._spn_units = Spinbox(self._frm_options,
                                  values=(UMM, UIK, UI, UMK),
                                  width=13, state=READONLY, readonlybackground=ENTCOL,
                                  wrap=True, textvariable=self._units)
        self._chk_scroll = Checkbutton(self._frm_options, text="Autoscroll",
                                      variable=self._autoscroll)
        self._spn_maxlines = Spinbox(self._frm_options, values=("100", "200", "500", "1000", "2000"),
                                    width=6, readonlybackground=ENTCOL, wrap=True,
                                    textvariable=self._maxlines, state=READONLY)
        self._chk_webmap = Checkbutton(self._frm_options, text="Web Map  Zoom",
                                      variable=self._webmap)
        self._scl_mapzoom = Scale(self._frm_options, from_=1, to=20, orient=HORIZONTAL,
                                 relief="sunken", bg=ENTCOL, variable=self._mapzoom)

    def _do_layout(self):
        '''
        Position widgets in frame.
        '''

        self._frm_basic.grid(column=0, row=0, columnspan=4, sticky=(W, E))
        self._lbl_port.grid(column=0, row=0, sticky=(W))
        self._lbx_port.grid(column=1, row=0, sticky=(W, E), padx=3, pady=3)
        self._scr_portv.grid(column=2, row=0, sticky=(N, S))
        self._scr_porth.grid(column=1, row=1, sticky=(E, W))
        self._lbl_baudrate.grid(column=0, row=2, sticky=(W))
        self._spn_baudrate.grid(column=1, row=2, sticky=(W), padx=3, pady=3)
        self._btn_toggle.grid(column=2, row=2, sticky=(E))

        self._frm_advanced.grid_forget()
        self._lbl_databits.grid(column=0, row=0, sticky=(W))
        self._spn_databits.grid(column=1, row=0, sticky=(W), padx=3, pady=3)
        self._lbl_stopbits.grid(column=2, row=0, sticky=(W))
        self._spn_stopbits.grid(column=3, row=0, sticky=(W), padx=3, pady=3)
        self._lbl_parity.grid(column=0, row=1, sticky=(W))
        self._spn_parity.grid(column=1, row=1, sticky=(W), padx=3, pady=3)
        self._chk_rts.grid(column=2, row=1, sticky=(W))
        self._chk_xon.grid(column=3, row=1, sticky=(W), padx=3, pady=3)

        ttk.Separator(self).grid(column=0, row=2, columnspan=4,
                                            padx=3, pady=3, sticky=(W, E))

        self._frm_buttons.grid(column=0, row=3, columnspan=4, sticky=(W, E))
        self._btn_connect.grid(column=0, row=0, padx=3, pady=3)
        self._btn_disconnect.grid(column=1, row=0, padx=3, pady=3)
        self._lbl_status.grid(column=2, row=0, padx=3, pady=3, sticky=(E))

        ttk.Separator(self).grid(column=0, row=4, columnspan=4,
                                            padx=3, pady=3, sticky=(W, E))

        self._frm_options.grid(column=0, row=5, columnspan=4, sticky=(W))
        self._lbl_protocol.grid(column=0, row=0, padx=3, pady=3, sticky=(W))
        self._rad_nmea.grid(column=1, row=0, padx=0, pady=0, sticky=(W))
        self._rad_ubx.grid(column=2, row=0, padx=0, pady=0, sticky=(W))
        self._rad_all.grid(column=3, row=0, padx=0, pady=0, sticky=(W))
        self._lbl_display.grid(column=0, row=1, padx=2, pady=3, sticky=(W))
        self._rad_parsed.grid(column=1, row=1, padx=2, pady=3, sticky=(W))
        self._rad_raw.grid(column=2, row=1, padx=2, pady=3, sticky=(W))
        self._lbl_format.grid(column=0, row=2, padx=3, pady=3, sticky=(W))
        self._spn_format.grid(column=1, row=2, padx=2, pady=3, sticky=(W))
        self._lbl_units.grid(column=0, row=3, padx=3, pady=3, sticky=(W))
        self._spn_units.grid(column=1, row=3, columnspan=3, padx=2, pady=3, sticky=(W))
        self._chk_scroll.grid(column=0, row=4, padx=3, pady=3, sticky=(W))
        self._spn_maxlines.grid(column=1, row=4, columnspan=3, padx=3, pady=3, sticky=(W))
        self._chk_webmap.grid(column=0, row=5, sticky=(W))
        self._scl_mapzoom.grid(column=1, row=5, columnspan=3, sticky=(W))

    def _on_select_port(self, *args, **kwargs):
        '''
        Get selected port from listbox and set global variable.
        '''

        idx = self._lbx_port.curselection()
        if  idx == "":
            idx = 0
        port_orig = self._lbx_port.get(idx)
        port = port_orig[0:port_orig.find(":")]
        desc = port_orig[port_orig.find(":") + 1:]
        if desc == '':
            desc = "device"
        self._port.set(port)
        self._port_desc.set(desc)

    def _toggle_advanced(self):
        '''
        Toggle advanced serial port settings panel on or off
        '''

        self._show_advanced = not self._show_advanced
        if self._show_advanced:
            self._frm_advanced.grid(column=0, row=1, columnspan=3, sticky=(W, E))
            self._btn_toggle.config(text=ADVON)
        else:
            self._frm_advanced.grid_forget()
            self._btn_toggle.config(text=ADVOFF)

    def _get_ports(self):
        '''
        Populate list of available serial ports using pyserial comports tool.
        If no ports found, disable all connection-dependent widgets.
        
        Attempt to preselect the first port that has a recognisable 
        GPS designation in its description (usually only works on 
        Posix platforms - Windows doesn't parse UART device desc or HWID)
        '''

        self._ports = sorted(comports())
        init_idx = 0
        port = ''
        desc = ''
        if len(self._ports) > 0:
            for idx, (port, desc, _) in enumerate(self._ports, 1):
                self._lbx_port.insert(idx, port + ": " + desc)
                for s in KNOWNGPS:
                    if s in desc:
                        init_idx = idx
                        break
            self._noports = False
        else:
            self._noports = True
            self.set_controls(NOPORTS)
        self._lbx_port.activate(init_idx)
        self._port.set(port)
        self._port_desc.set(desc)

    def _reset(self):
        '''
        Reset settings to defaults.
        '''

        self._baudrate.set(BAUDRATES[4])  # 9600
        self._databits.set(8)
        self._stopbits.set(1)
        self._parity.set("None")
        self._rtscts.set(False)
        self._xonxoff.set(False)
        self._protocol.set(MIXED_PROTOCOL)
        self._format.set(DDD)
        self._units.set(UMM)
        self._autoscroll.set(1)
        self._maxlines.set(300)
        self._raw.set(False)
        self._webmap.set(False)
        self._mapzoom.set(10)

    def _save_settings(self):
        '''
        TODO Save settings.
        '''

        # print("Save settings clicked")

    def _load_settings(self):
        '''
        TODO Load settings. TODO
        '''

        # print("Load settings clicked")

    def set_controls(self, status):
        '''
        Public method to enable and disable serial port controls
        depending on connection status.
        '''

        self._lbl_port.configure(state=(NORMAL if status == DISCONNECTED else DISABLED))
        self._lbx_port.configure(state=(NORMAL if status == DISCONNECTED else DISABLED))
        self._lbl_baudrate.configure(state=(NORMAL if status == DISCONNECTED else DISABLED))
        self._spn_baudrate.configure(state=(READONLY if status == DISCONNECTED else DISABLED))
        self._lbl_databits.configure(state=(NORMAL if status == DISCONNECTED else DISABLED))
        self._spn_databits.configure(state=(READONLY if status == DISCONNECTED else DISABLED))
        self._lbl_stopbits.configure(state=(NORMAL if status == DISCONNECTED else DISABLED))
        self._spn_stopbits.configure(state=(READONLY if status == DISCONNECTED else DISABLED))
        self._lbl_parity.configure(state=(NORMAL if status == DISCONNECTED else DISABLED))
        self._spn_parity.configure(state=(READONLY if status == DISCONNECTED else DISABLED))
        self._chk_rts.configure(state=(NORMAL if status == DISCONNECTED else DISABLED))
        self._chk_xon.configure(state=(NORMAL if status == DISCONNECTED else DISABLED))
        self._btn_connect.config(state=(DISABLED if status in (CONNECTED, NOPORTS) else NORMAL))
        self._btn_disconnect.config(state=(DISABLED if status in (DISCONNECTED, NOPORTS) else NORMAL))

#         self._rad_nmea.config(state=(DISABLED if status in (CONNECTED, NOPORTS) else NORMAL))
#         self._rad_ubx.config(state=(DISABLED if status in (CONNECTED, NOPORTS) else NORMAL))
#         self._rad_all.config(state=(DISABLED if status in (CONNECTED, NOPORTS) else NORMAL))

    def get_settings(self):
        '''
        Public method returns all settings as a dict.
        '''

        self._settings['port'] = self._port.get()
        self._settings['noports'] = self._noports
        self._settings['port_desc'] = self._port_desc.get()
        self._settings['baudrate'] = self._baudrate.get()
        self._settings['databits'] = self._databits.get()
        self._settings['stopbits'] = self._stopbits.get()
        self._settings['parity'] = self._parity.get()
        self._settings['rtscts'] = self._rtscts.get()
        self._settings['xonxoff'] = self._xonxoff.get()
        self._settings['protocol'] = self._protocol.get()
        self._settings['raw'] = self._raw.get()
        self._settings['autoscroll'] = self._autoscroll.get()
        self._settings['maxlines'] = self._maxlines.get()
        self._settings['webmap'] = self._webmap.get()
        self._settings['mapzoom'] = self._mapzoom.get()
        self._settings['units'] = self._units.get()
        self._settings['format'] = self._format.get()
        return self._settings

    def get_size(self):
        '''
        Get current frame size.
        '''

        self.update_idletasks()  # Make sure we know about any resizing
        return (self.winfo_width(), self.winfo_height())



'''
Mapview frame class for PyGPSClient application.

This handles a frame containing a location map downloaded via a MapQuest API.

*****************************************************************************
NOTE: The free MapQuest API key is subject to a limit of 15,000 
transactions / month, or roughly 500 / day, so the map updates are only
run periodically (once a minute). This utility is NOT intended to be used for
real time navigation.
*****************************************************************************

Created on 13 Sep 2020

@author: semuadmin
'''

from io import BytesIO
from time import time
from tkinter import Frame, Canvas, font, NW, N, S, E, W

from PIL import ImageTk, Image
import requests

from .globals import WIDGETU2, MAPURL, MAP_UPDATE_INTERVAL, \
                                IMG_WORLD, ICON_POS
from .strings import NOWEBMAPERROR1, NOWEBMAPERROR2, NOWEBMAPERROR3, \
                                NOWEBMAPERROR4, NOWEBMAPERROR5


class MapviewFrame(Frame):
    '''
    Frame inheritance class for plotting satellite view.
    '''

    def __init__(self, app, *args, **kwargs):
        '''
        Constructor.
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)

        Frame.__init__(self, self.__master, *args, **kwargs)

        def_w, def_h = WIDGETU2
        self.width = kwargs.get('width', def_w)
        self.height = kwargs.get('height', def_h)
        self._img = None
        self._marker = None
        self._last_map_update = 0
        self._body()

        self.bind("<Configure>", self._on_resize)

    def _body(self):
        '''
        Set up frame and widgets.
        '''

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.can_mapview = Canvas(self, width=self.width, height=self.height, bg="white")
        self.can_mapview.grid(column=0, row=0, sticky=(N, S, E, W))

    def update_map(self, **kwargs):
        '''
        Draw map and mark current known position.
        '''

        if 'lat' in kwargs and 'lon' in kwargs:
            lat = kwargs['lat']
            lon = kwargs['lon']
        else:
            return
        vacc = kwargs.get('vacc', 0)
        hacc = kwargs.get('hacc', 0)
        fix = kwargs.get('fix', 'NO-FIX')
        static = kwargs.get('static', True)

        w, h = self.width, self.height
        resize_font = font.Font(size=min(int(w / 20), 14))

        if fix == 'NO FIX' or lat == '' or lon == '':
            self.can_mapview.delete("all")
            self.reset_map_refresh()
            self.can_mapview.create_text(w / 2, h / 2 - (w / 20), text=NOWEBMAPERROR1,
                                         fill="red", font=resize_font)
            self.can_mapview.create_text(w / 2, h / 2 + (w / 20), text=NOWEBMAPERROR5,
                                         fill="red", font=resize_font)
            return

        if static:
            self._draw_static_map(lat, lon)
        else:
            self._draw_web_map(lat, lon, hacc, vacc)

    def _draw_static_map(self, lat, lon):
        '''
        Draw fixed scale Mercator world map
        '''

        OFFSET_X = 0
        OFFSET_Y = 0

        w, h = self.width, self.height
        self.can_mapview.delete("all")
        self._img = ImageTk.PhotoImage(Image.open(IMG_WORLD).resize((w, h), Image.ANTIALIAS))
        self._marker = ImageTk.PhotoImage(Image.open(ICON_POS))
        self.can_mapview.create_image(0, 0, image=self._img, anchor=NW)
        x = (w / 2) + int((lon * (w / 360))) + OFFSET_X
        y = (h / 2) - int((lat * (h / 180))) + OFFSET_Y
        self.can_mapview.create_image(x, y, image=self._marker, anchor=S)

    def _draw_web_map(self, lat, lon, hacc, vacc):
        '''
        Draw scalable web map via MapQuest API
        '''

        w, h = self.width, self.height
        resize_font = font.Font(size=min(int(w / 20), 14))

        apikey = self.__app.api_key
        if apikey == '':
            self.can_mapview.delete("all")
            self.reset_map_refresh()
            self.can_mapview.create_text(w / 2, h / 2 - (w / 20), text=NOWEBMAPERROR1,
                                         fill="orange", font=resize_font)
            self.can_mapview.create_text(w / 2, h / 2 + (w / 20), text=NOWEBMAPERROR2,
                                         fill="orange", font=resize_font)
            return

        now = time()
        if now - self._last_map_update < MAP_UPDATE_INTERVAL:
            self._draw_countdown((-360 / MAP_UPDATE_INTERVAL) * (now - self._last_map_update))
            return
        else:
            self._last_map_update = now

        url = self._format_url(apikey, lat, lon, hacc, vacc)

        try:
            response = requests.get(url)
            if response.status_code == 200:
                img_data = response.content
                self._img = ImageTk.PhotoImage(Image.open(BytesIO(img_data)))
                self.can_mapview.delete("all")
                self.can_mapview.create_image(0, 0, image=self._img, anchor=NW)
                self.can_mapview.update()
        except Exception:  # Complex exception chain with these calls so catch everything
            self.can_mapview.delete("all")
            self.can_mapview.create_text(w / 2, h / 2 - (w / 20), text=NOWEBMAPERROR3,
                                         fill="red", font=resize_font)
            self.can_mapview.create_text(w / 2, h / 2 + (w / 20), text=NOWEBMAPERROR4,
                                         fill="red", font=resize_font)

    def _draw_countdown(self, wait):
        '''
        Draw clock icon indicating time until next scheduled map refresh.
        '''

        self.can_mapview.create_oval((5, 5, 20, 20), fill="#616161", outline="")
        self.can_mapview.create_arc((5, 5, 20, 20), start=90, extent=wait, \
                                            fill="#ffffff", outline="")
        self.can_mapview.update()

    def _format_url(self, apikey, lat, lon, hacc, vacc):
        '''
        Formats URL for web map download.
        '''

        w, h = self.width, self.height
        zoom = self.__app.frm_settings.get_settings()['mapzoom']
        radius = str(hacc / 1000)  # km

        return MAPURL.format(apikey, lat, lon, zoom, radius, lat, lon, w, h)

    def reset_map_refresh(self):
        '''
        Reset map refresh counter to zero
        '''

        self._last_map_update = 0

    def _on_resize(self, event):
        '''
        Resize frame
        '''

        self.width, self.height = self.get_size()

    def get_size(self):
        '''
        Get current canvas size.
        '''

        self.update_idletasks()  # Make sure we know about any resizing
        width = self.can_mapview.winfo_width()
        height = self.can_mapview.winfo_height()
        return (width, height)



#!/usr/bin/env python
'''
Setup script for PyGPSClient Application

python setup.py sdist bdist_wheel

Created on 12 Sep 2020

@author: semuadmin
'''

from setuptools import setup, find_packages

from pygpsclient._version import VERSION

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="PyGPSClient",
    version=VERSION,
    packages=find_packages(),
    install_requires=["pyubx2>=0.1.7", "pynmea2>=1.15.0", "requests>=2.24.0",
                      "Pillow>=7.2.0", "pyserial>=3.4"],
    package_data={
        "pygpsclient": ["resources/*.gif", "resources/*.png",
                     "resources/*.ico", "resources/*.icns",
                     "images/*.png"],
    },
    include_package_data=True,
    author="semuadmin",
    author_email="semuadmin@semuconsulting.com",
    description="PyGPSClient GPS Client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/semuconsulting/PyGPSClient",
    keywords="PyGPSClient GPS NMEA UBX",
    platforms="Windows, MacOS, Linux",
    license="BSD 3-Clause 'Modified' License",
    project_urls={
        "Bug Tracker": "https://github.com/semuconsulting/PyGPSClient",
        "Documentation": "https://github.com/semuconsulting/PyGPSClient",
        "Source Code": "https://github.com/semuconsulting/PyGPSClient",
    },
    classifiers=[
        "License :: OSI Approved :: BSD License",
        'Development Status :: 3 - Alpha',
        'Environment :: MacOS X',
        'Environment :: X11 Applications',
        'Environment :: Win32 (MS Windows)',
        'Environment :: Console',
        'Intended Audience :: End Users/Desktop',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3',
        'Topic :: Desktop Environment',
        'Topic :: Terminals :: Serial',
    ]

)



'''
Satview frame class for PyGPSClient application.

This handles a frame containing a 2D plot of satellite visibility.

Created on 13 Sep 2020

@author: semuadmin
'''

from tkinter import Frame, Canvas, font, BOTH, YES
from operator import itemgetter
from .globals import snr2col, cel2cart, WIDGETU1, BGCOL, FGCOL


def _create_circle(self, x, y, r, **kwargs):
    return self.create_oval(x - r, y - r, x + r, y + r, **kwargs)


Canvas.create_circle = _create_circle


class SkyviewFrame(Frame):
    '''
    Frame inheritance class for plotting satellite view.
    '''

    def __init__(self, app, *args, **kwargs):
        '''
        Constructor.
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)

        Frame.__init__(self, self.__master, *args, **kwargs)

        def_w, def_h = WIDGETU1
        self.width = kwargs.get('width', def_w)
        self.height = kwargs.get('height', def_h)
        self.bg_col = BGCOL
        self.fg_col = FGCOL
        self.body()

        self.bind("<Configure>", self._on_resize)

    def body(self):
        '''
        Set up frame and widgets.
        '''

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self.can_satview = Canvas(self, width=self.width, height=self.height,
                                  bg=self.bg_col)
        self.can_satview.pack(fill=BOTH, expand=YES)

    def init_sats(self):
        '''
        Initialise satellite view
        '''

        w, h = self.width, self.height
        axis_r = w / 15
        resize_font = font.Font(size=min(int(w / 25), 10))
        self.can_satview.delete("all")
        maxr = min((h / 2), (w / 2)) - (axis_r * 2)
        for r in (.2, .4, .6, .8, 1):
            self.can_satview.create_circle(w / 2, h / 2, maxr * r, outline=self.fg_col, width=1)
        self.can_satview.create_line(w / 2, 0, w / 2 , h, fill=self.fg_col)
        self.can_satview.create_line(0, h / 2, w, h / 2, fill=self.fg_col)
        self.can_satview.create_text(w - axis_r, h / 2, text="90\u00b0\n E", fill=self.fg_col,
                                     font=resize_font)
        self.can_satview.create_text(axis_r, h / 2, text="270\u00b0\n W", fill=self.fg_col,
                                     font=resize_font)
        self.can_satview.create_text(w / 2, axis_r, text="0\u00b0 N", fill=self.fg_col,
                                     font=resize_font)
        self.can_satview.create_text(w / 2, h - axis_r, text="180\u00b0 S", fill=self.fg_col,
                                     font=resize_font)

    def update_sats(self, data):
        '''
        Plot satellites' elevation and azimuth position.
        '''

        w, h = self.width, self.height
        axis_r = w / 15
        maxr = min((h / 2), (w / 2)) - (axis_r * 2)
        resize_font = font.Font(size=min(int(maxr / 10), 10))
        self.init_sats()

        for d in sorted(data, key=itemgetter(3)):  # sort by ascending snr
            try:
                prn, ele, azi, snr = d
                ele = int(ele)
                azi = (int(azi) - 90) % 360  # adjust so north is upwards
                x, y = cel2cart(ele, azi)
                x = x * maxr
                y = y * maxr
                if snr == '':
                    snr = 0
                else:
                    snr = int(snr)
                if int(prn) > 96:  # OTHER e.g. GAL
                    ol_col = "grey"
                elif 65 <= int(prn) <= 96:  # GLONASS
                    ol_col = "brown"
                elif 33 <= int(prn) <= 64:  # SBAS
                    ol_col = "blue"
                else:  # original GPS
                    ol_col = "black"
                prn = f"{int(prn):02}"
                self.can_satview.create_circle(x + (w / 2), y + (h / 2),
                                               (maxr / 10), outline=ol_col,
                                               fill=snr2col(snr))
                self.can_satview.create_text(x + (w / 2), y + (h / 2), text=prn,
                                             fill=self.fg_col, font=resize_font)
            except ValueError:
                pass

        self.can_satview.update()

    def _on_resize(self, event):
        '''
        Resize frame
        '''

        self.width, self.height = self.get_size()

    def get_size(self):
        '''
        Get current canvas size.
        '''

        self.update_idletasks()  # Make sure we know about any resizing
        width = self.can_satview.winfo_width()
        height = self.can_satview.winfo_height()
        return (width, height)



'''
NMEA Protocol handler - handles all incoming standard and proprietary NMEA sentences

Uses pynmea2 library for parsing

Created on 30 Sep 2020

@author: semuadmin
'''

# pylint: disable=invalid-name
from time import time
from pynmea2 import parse as nmea_parse, ParseError, types, RMC, VTG, GSV, GGA, GSA, GLL
from pynmea2.types.proprietary.ubx import UBX00

from .globals import DEVICE_ACCURACY, HDOP_RATIO, SAT_EXPIRY
from .strings import NMEAVALERROR


class NMEAHandler():
    '''
    NMEAHandler class
    '''

    def __init__(self, app):
        '''
        Constructor.
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)

        self._raw_data = None
        self._parsed_data = None
        self.gsv_data = []  # Holds array of current satellites in view from NMEA GSV sentences
        self.gsv_log = {}  # Holds cumulative log of all satellites seen
        self.lon = 0
        self.lat = 0
        self.alt = 0
        self.track = 0
        self.speed = 0
        self.pdop = 0
        self.hdop = 0
        self.vdop = 0
        self.hacc = 0
        self.vacc = 0
        self.utc = ''
        self.sip = 0
        self.fix = '-'

    def process_data(self, data: bytes):
        '''
        Process NMEA message type
        '''

        try:
            parsed_data = nmea_parse(data)
        except ParseError:
            # Parsing errors at this point are typically due to NMEA and UBX
            # protocols getting garbled in the input stream. It only happens
            # rarely so we ignore them and carry on.
            return None

        if data or parsed_data:
            self._update_console(data, parsed_data)

        if isinstance(parsed_data, RMC):  # Recommended minimum data for GPS
            self._process_RMC(parsed_data)
        if isinstance(parsed_data, GGA):  # GPS Fix Data
            self._process_GGA(parsed_data)
        if isinstance(parsed_data, GLL):  # GLL Lat Lon Data
            self._process_GLL(parsed_data)
        if isinstance(parsed_data, GSA):  # GPS DOP (Dilution of Precision) and active satellites
            self._process_GSA(parsed_data)
        if isinstance(parsed_data, VTG):  # GPS Vector track and Speed over the Ground
            self._process_VTG(parsed_data)
        if isinstance(parsed_data, GSV):  # GPS Satellites in View
            self._process_GSV(parsed_data)
        if isinstance(parsed_data, UBX00):  # GPS Satellites in View
            self._process_UBX00(parsed_data)

        return self._parsed_data

    def _update_console(self, raw_data, parsed_data):
        '''
        Write the incoming data to the console in raw or parsed format.
        '''

        if self.__app.frm_settings.get_settings()['raw']:
            self.__app.frm_console.update_console(repr(raw_data))
        else:
            self.__app.frm_console.update_console(repr(parsed_data))

    def _process_RMC(self, data: types.talker):
        '''
        Process RMC sentence - Recommended minimum data for GPS.
        '''

        try:
            self.utc = data.timestamp
            (self.lat, self.lon) = self.nmea2latlon(data)
            self.speed = data.spd_over_grnd
            self.track = data.true_course
            self.__app.frm_banner.update_banner(time=self.utc, lat=self.lat,
                                                lon=self.lon, speed=self.speed,
                                                track=self.track)
            if self.__app.frm_settings.get_settings()['webmap']:
                self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, static=False)
            else:
                self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, static=True)
        except ValueError as err:
            self.__app.set_status(NMEAVALERROR.format(err), "red")

    def _process_GGA(self, data: types.talker):
        '''
        Process GGA sentence - GPS Fix Data.
        '''

        try:
            self.utc = data.timestamp
            self.sip = data.num_sats
            (self.lat, self.lon) = self.nmea2latlon(data)
            self.alt = data.altitude
            self.__app.frm_banner.update_banner(time=self.utc, lat=self.lat,
                                                lon=self.lon, alt=self.alt, sip=self.sip)
            if self.__app.frm_settings.get_settings()['webmap']:
                self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, static=False)
            else:
                self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, static=True)
        except ValueError as err:
            self.__app.set_status(NMEAVALERROR.format(err), "red")

    def _process_GLL(self, data: types.talker):
        '''
        Process GLL sentence - GPS Lat Lon.
        '''

        try:
            self.utc = data.timestamp
            (self.lat, self.lon) = self.nmea2latlon(data)
            self.__app.frm_banner.update_banner(time=self.utc, lat=self.lat, lon=self.lon)
            if self.__app.frm_settings.get_settings()['webmap']:
                self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, static=False)
            else:
                self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, static=True)
        except ValueError as err:
            self.__app.set_status(NMEAVALERROR.format(err), "red")

    def _process_GSA(self, data: types.talker):
        '''
        Process GSA sentence - GPS DOP (Dilution of Precision) and active satellites.
        '''

        self.pdop = float(data.pdop)
        self.hdop = float(data.hdop)
        self.vdop = float(data.vdop)
        if data.mode_fix_type == '3':
            fix = '3D'
        elif data.mode_fix_type == '2':
            fix = '2D'
        else:
            fix = 'NO FIX'

        if self.__app.frm_settings.get_settings()['webmap']:
            self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, hacc=self.hacc,
                                              vacc=self.vacc, fix=fix, static=False)
        else:
            self.__app.frm_mapview.update_map(lat=self.lat, lon=self.lon, hacc=self.hacc,
                                              vacc=self.vacc, fix=fix, static=True)

        self.__app.frm_banner.update_banner(dop=self.pdop, hdop=self.hdop, vdop=self.vdop, fix=fix)

    def _process_GSV(self, data: types.talker):
        '''
        Process GSV sentences - GPS Satellites in View
        These come in batches of 1-4 sentences, each containing the positions
        of up to 4 satellites (16 satellites in total).
        Modern receivers can send multiple batches corresponding to different
        NMEA assigned 'ID' ranges (GPS 1-32, SBAS 33-64, GLONASS 65-96)
        '''

        self.gsv_data = []
        gsv_dict = {}
        now = time()

        try:
            if data.sv_prn_num_1 != '':
                gsv_dict[data.sv_prn_num_1] = (data.elevation_deg_1,
                                               data.azimuth_1, data.snr_1, now)
            if data.sv_prn_num_2 != '':
                gsv_dict[data.sv_prn_num_2] = (data.elevation_deg_2,
                                               data.azimuth_2, data.snr_2, now)
            if data.sv_prn_num_3 != '':
                gsv_dict[data.sv_prn_num_3] = (data.elevation_deg_3,
                                               data.azimuth_3, data.snr_3, now)
            if data.sv_prn_num_4 != '':
                gsv_dict[data.sv_prn_num_4] = (data.elevation_deg_4,
                                               data.azimuth_4, data.snr_4, now)
        except AttributeError:
            pass

        for key in gsv_dict:
            self.gsv_log[key] = gsv_dict[key]

        for key in self.gsv_log:
            elev, azim, snr, lastupdate = self.gsv_log[key]
            if now - lastupdate < SAT_EXPIRY:  # expire passed satellites
                self.gsv_data.append((key, elev, azim, snr))

        self.__app.frm_satview.update_sats(self.gsv_data)
        self.__app.frm_banner.update_banner(siv=len(self.gsv_data))
        self.__app.frm_graphview.update_graph(self.gsv_data, len(self.gsv_data))

    def _process_VTG(self, data: types.talker):
        '''
        Process VTG sentence - GPS Vector track and Speed over the Ground.
        '''

        try:
            self.track = data.true_track
            self.speed = data.spd_over_grnd_kmph
            self.__app.frm_banner.update_banner(speed=self.speed, track=self.track)
        except ValueError as err:
            self.__app.set_status(NMEAVALERROR.format(err), "red")

    def _process_UBX00(self, data: types.ubx.UBX00):
        '''
        Process UXB00 sentence - GPS Vector track and Speed over the Ground.
        '''

        try:
            self.hacc = float(data.h_acc)
            self.vacc = float(data.v_acc)
            self.__app.frm_banner.update_banner(hacc=self.hacc, vacc=self.vacc)
        except ValueError as err:
            self.__app.set_status(NMEAVALERROR.format(err), "red")

    @staticmethod
    def nmea2latlon(data: types.talker) -> (float, float):
        '''
        Convert parsed NMEA GGA sentence to decimal lat, lon
        '''

        if data.lat == '':
            lat = ''
        else:
            latdeg = float(data.lat[0:2])
            latmin = float(data.lat[2:])
            londeg = float(data.lon[0:3])
            lat = (latdeg + latmin / 60) * (-1 if data.lat_dir == 'S' else 1)
        if data.lon == '':
            lon = ''
        else:
            lonmin = float(data.lon[3:])
            lon = (londeg + lonmin / 60) * (-1 if data.lon_dir == 'W' else 1)
        return (lat, lon)

    @staticmethod
    def _estimate_acc(dop) -> float:
        '''
        Derive a graphic indication of positional accuracy (in m) based on the HDOP
        (Horizontal Dilution of Precision) value and the nominal native device
        accuracy (datasheet CEP)

        NB: this is a largely arbitrary estimate - there is no direct correlation
        between HDOP and accuracy based solely on generic NMEA data.
        The NMEA PUBX,00 or UBX NAV-POSLLH message types return an explicit estimate
        of horizontal and vertical accuracy and are the preferred source.
        '''

        return float(dop) * DEVICE_ACCURACY * HDOP_RATIO / 1000



'''
Menubar class for PyGPSClient application.

This handles the menu bar.

Created on 12 Sep 2020

@author: semuadmin
'''

from tkinter import Menu

from .strings import MENUFILE, MENUVIEW, MENUOPTION, MENUEXIT, \
                                MENUHIDESE, MENUHIDESB, MENUHIDECON, MENUHIDEMAP, \
                                MENUHIDESATS, MENUABOUT, MENUHELP, MENUUBXCONFIG


class MenuBar(Menu):
    '''
    Menu inheritance class for menu bar.
    '''

    def __init__(self, app, *args, **kwargs):
        '''
        Constructor
        '''

        self.__app = app  # Reference to main application class
        self.__master = self.__app.get_master()  # Reference to root class (Tk)
        Menu.__init__(self, self.__master, *args, **kwargs)

        self.option_add("*Font", 'TkMenuFont')
        # Create a pull-down menu for file operations
        self.file_menu = Menu(self, tearoff=False)
#         self.file_menu.add_command(label=MENUSAVE, underline=1,
#                                    command=self.__app.frm_settings._save_settings)
#         self.file_menu.add_command(label=MENULOAD, underline=1,
#                                    command=self.__app.frm_settings._load_settings)
        self.file_menu.add_command(label=MENUEXIT, underline=1, accelerator="Ctrl-Q",
                                   command=self.__app.exit)
        self.add_cascade(menu=self.file_menu, label=MENUFILE)

        # Create a pull-down menu for view operations
        self.view_menu = Menu(self, tearoff=False)
        self.view_menu.add_command(label=MENUHIDESE, underline=1,
                                     command=self.__app.toggle_settings)
        self.view_menu.add_command(label=MENUHIDESB, underline=1,
                                     command=self.__app.toggle_status)
        self.view_menu.add_command(label=MENUHIDECON, underline=1,
                                     command=self.__app.toggle_console)
        self.view_menu.add_command(label=MENUHIDEMAP, underline=1,
                                   command=self.__app.toggle_map)
        self.view_menu.add_command(label=MENUHIDESATS, underline=1,
                                   command=self.__app.toggle_sats)
        self.add_cascade(menu=self.view_menu, label=MENUVIEW)

        # Create a pull-down menu for view operations
        self.options_menu = Menu(self, tearoff=False)
        self.options_menu.add_command(label=MENUUBXCONFIG, underline=1,
                                   command=self.__app.ubxconfig)
        self.add_cascade(menu=self.options_menu, label=MENUOPTION)

        # Create a pull-down menu for help operations
        self.help_menu = Menu(self, tearoff=False)
        self.help_menu.add_command(label=MENUABOUT, underline=1,
                                   command=self.__app.about)
        self.add_cascade(menu=self.help_menu, label=MENUHELP)
